"""
Created on May 17 2020
@author: Pedram Tavadze
"""
import re

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from .abinitparser import AbinitParser
from .doscarplot import DosPlot
from .elkparser import ElkParser
from .procarparser import ProcarParser
from .procarplot import ProcarPlot
from .procarselect import ProcarSelect
from .splash import welcome
from .utilsprocar import UtilsProcar

plt.rcParams["mathtext.default"] = "regular"  # Roman ['rm', 'cal', 'it', 'tt', 'sf',
#                                                   'bf', 'default', 'bb', 'frak',
#                                                   'circled', 'scr', 'regular']
plt.rcParams["font.family"] = "Georgia"
plt.rc("font", size=18)  # controls default text sizes
plt.rc("axes", titlesize=22)  # fontsize of the axes title
plt.rc("axes", labelsize=22)  # fontsize of the x and y labels
plt.rc("xtick", labelsize=22)  # fontsize of the tick labels
plt.rc("ytick", labelsize=22)  # fontsize of the tick labels
plt.rc("legend", fontsize=18)  # legend fontsize
# plt.rc('figure', titlesize=22)  # fontsize of the figure title


def bandsdosplot(
    bands_file="PROCAR",
    dos_file="vasprun.xml",
    outcar="OUTCAR",
    abinit_output=None,
    bands_mode="plain",
    dos_mode="plain",
    plot_total=True,
    fermi=None,
    mask=None,
    markersize=0.02,
    marker="o",
    atoms=None,
    orbitals=None,
    bands_spin=0,
    bands_separate=False,
    dos_spins=None,
    dos_labels=None,
    dos_spin_colors=[(1, 0, 0), (0, 0, 1)],
    dos_colors=None,
    dos_items=None,
    elimit=None,
    vmin=None,
    vmax=None,
    cmap="jet",
    grid=False,
    kpointsfile=None,
    code="vasp",
    savefig=None,
    title=None,
    kdirect=True,
    discontinuities=None,
):
    """This function creates plots containing both DOS and bands."""

    welcome()

    fig = plt.figure(figsize=(13, 7), constrained_layout=False)
    widths = [13, 5]
    heights = [9]
    gs = fig.add_gridspec(1, 2, width_ratios=widths, height_ratios=heights)

    ax1 = fig.add_subplot(gs[0, 0])
    ax2 = fig.add_subplot(gs[0, 1])

    total = plot_total

    if atoms is None:
        bands_atoms = [-1]
    else:
        bands_atoms = atoms.copy()
    if orbitals is None:
        bands_orbitals = [-1]
    else:
        bands_orbitals = (
            orbitals.copy()
        )  # these copying is because orbitals select changes shifts orbitals by one

    if dos_spins is None:
        dos_spins = [0, 1]

    # Verbose section

    print("Script initiated")
    print("code           : ", code)
    print("bands file     : ", bands_file)
    print("bands mode     : ", bands_mode)
    print("bands spin     : ", bands_spin)
    print("dos file       : ", dos_file)
    print("dos mode       : ", dos_mode)
    print("dos spin       : ", dos_spins)
    print("atoms list     : ", atoms)
    print("orbs. list     : ", orbitals)

    if fermi is None and outcar is None and abinit_output is None:
        print(
            "WARNING : Fermi Energy not set! Please set manually or provide output file and set code type."
        )
        fermi = 0

    print("fermi energy   : ", fermi)
    print("energy range   : ", elimit)

    if mask is not None:
        print("masking thres. : ", mask)

    print("colormap       : ", cmap)
    print("markersize     : ", markersize)
    print("vmax           : ", vmax)
    print("vmin           : ", vmin)
    print("grid enabled   : ", grid)
    print("savefig        : ", savefig)
    print("title          : ", title)
    print("outcar         : ", outcar)

    if kdirect:
        print("k-grid         : reduced")
    else:
        print(
            "k-grid         : cartesian (Remember to provide an output file for this case to work.)"
        )

    if discontinuities is None:
        discontinuities = []

    #### READING KPOINTS FILE IF PRESENT ####

    # If KPOINTS file is given:
    if kpointsfile is not None:
        # Getting the high symmetry point names from KPOINTS file
        f = open(kpointsfile)
        KPread = f.read()
        f.close()

        KPmatrix = re.findall("reciprocal[\s\S]*", KPread)
        tick_labels = np.array(re.findall("!\s(.*)", KPmatrix[0]))
        knames = []
        knames = [tick_labels[0]]

        ################## Checking for discontinuities ########################
        discont_indx = []
        icounter = 1
        while icounter < len(tick_labels) - 1:
            if tick_labels[icounter] == tick_labels[icounter + 1]:
                knames.append(tick_labels[icounter])
                icounter = icounter + 2
            else:
                discont_indx.append(icounter)
                knames.append(tick_labels[icounter] + "|" + tick_labels[icounter + 1])
                icounter = icounter + 2
        knames.append(tick_labels[-1])
        discont_indx = list(dict.fromkeys(discont_indx))

        ################# End of discontinuity check ##########################

        # Added by Nicholas Pike to modify the output of seekpath to allow for
        # latex rendering.
        for i in range(len(knames)):
            if knames[i] == "GAMMA":
                knames[i] = "\Gamma"
            else:
                pass

        knames = [str("$" + latx + "$") for latx in knames]

        # getting the number of grid points from the KPOINTS file
        f2 = open(kpointsfile)
        KPreadlines = f2.readlines()
        f2.close()
        numgridpoints = int(KPreadlines[1].split()[0])

        kticks = [0]
        gridpoint = 0
        for kt in range(len(knames) - 1):
            gridpoint = gridpoint + numgridpoints
            kticks.append(gridpoint - 1)

        print("knames         : ", knames)
        print("kticks         : ", kticks)

        # creating an array for discontunuity k-points. These are the indexes
        # of the discontinuity k-points.
        for k in discont_indx:
            discontinuities.append(kticks[int(k / 2) + 1])
        if discontinuities:
            print("discont. list  : ", discontinuities)

    #### END OF KPOINTS FILE DEPENDENT SECTION ####

    # spin = {"0": 0, "1": 1, "2": 2, "3": 3, "st": "st"}[str(spin)]

    #### parsing the PROCAR file or equivalent to retrieve spd data ####
    code = code.lower()
    if code == "vasp":
        procarFile = ProcarParser()
        dos_plot = DosPlot(dos_file)
        if dos_spins is None:
            dos_spins = np.arange(dos_plot.VaspXML.dos_total.ncols)

    # If ticks and names are given by the user manually:
    if kticks is not None and knames is not None:
        ticks = list(zip(kticks, knames))
    elif kticks is not None:
        ticks = list(zip(kticks, kticks))
    else:
        ticks = None

    if kpointsfile is None and kticks and knames:
        print("knames         : ", knames)
        print("kticks         : ", kticks)
        if discontinuities:
            print("discont. list  : ", discontinuities)

    # The second part of this function is parse/select/use the data in
    # OUTCAR (if given) and PROCAR

    # first parse the outcar if given, to get Efermi and Reciprocal lattice
    recLat = None
    if code == "vasp":
        if outcar:
            outcarparser = UtilsProcar()
            if fermi is None:
                fermi = outcarparser.FermiOutcar(outcar)
                print("Fermi energy   :  %s eV (from OUTCAR)" % str(fermi))
            recLat = outcarparser.RecLatOutcar(outcar)

    # if kdirect = False, then the k-points will be in cartesian coordinates.
    # The output should be read to find the reciprocal lattice vectors to transform
    # from direct to cartesian

    if code == "vasp":
        if kdirect:
            procarFile.readFile(bands_file, permissive=False)
        else:
            procarFile.readFile(bands_file, permissive=False, recLattice=recLat)

    # processing the data, getting an instance of the class that reduces the data
    data = ProcarSelect(procarFile, deepCopy=True, mode=bands_mode)
    numofbands = int(data.spd.shape[1] / 2)

    # handling the spin, `spin='st'` is not straightforward, needs
    # to calculate the k vector and its normal. Other `spin` values
    # are trivial.

    if bands_spin == "st":
        # two `ProcarSelect` instances, to store temporal values: spin_x, spin_y
        dataX = ProcarSelect(procarFile, deepCopy=True)
        dataX.selectIspin([1])
        dataX.selectAtoms(bands_atoms, fortran=False)

        dataX.selectOrbital(bands_orbitals)
        dataY = ProcarSelect(procarFile, deepCopy=True)
        dataY.selectIspin([2])
        dataY.selectAtoms(bands_atoms, fortran=False)
        dataY.selectOrbital(bands_orbitals)
        # getting the signed angle of each K-vector
        angle = np.arctan2(dataX.kpoints[:, 1], (dataX.kpoints[:, 0] + 0.000000001))
        sin = np.sin(angle)
        cos = np.cos(angle)
        sin.shape = (sin.shape[0], 1)
        cos.shape = (cos.shape[0], 1)
        # print sin, cos
        # storing the spin projection into the original array
        data.spd = -sin * dataX.spd + cos * dataY.spd
    else:
        data.selectIspin([bands_spin], separate=bands_separate)
        data.selectAtoms(bands_atoms, fortran=False)
        data.selectOrbital(bands_orbitals)

    # Plotting the data
    if bands_separate:
        if bands_spin == 0:
            # plotting spin up bands separately
            data.bands = (
                data.bands[:, :numofbands].transpose() - np.array(fermi)
            ).transpose()
            print("Plotting spin up bands...")

        elif bands_spin == 1:
            # plotting spin down bands separately
            data.bands = (
                data.bands[:, numofbands:].transpose() - np.array(fermi)
            ).transpose()
            print("Plotting spin down bands...")

        plot = ProcarPlot(data.bands, data.spd, data.kpoints, ax=ax1)

    else:
        # Regular plotting method. For spin it plots density or magnetization.
        data.bands = (data.bands.transpose() - np.array(fermi)).transpose()
        plot = ProcarPlot(data.bands, data.spd, data.kpoints)

    if vmin is None:
        vmin = plot.spd.min()
    if vmax is None:
        vmax = plot.spd.max()

    ###### start of mode dependent options #########
    if bands_mode == "scatter":
        _, ax1 = plot.scatterPlot(
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            ticks=ticks,
            discontinuities=discontinuities,
            ax=ax1,
            mask=mask,
        )
        if fermi is not None:
            ax1.set_ylabel(r"$E-E_f$ [eV]")
        else:
            ax1.set_ylabel(r"Energy [eV]")
        if elimit is not None:
            ax1.set_ylim(elimit)

    elif bands_mode == "plain":
        _, ax1 = plot.plotBands(ticks=ticks, discontinuities=discontinuities, ax=ax1,)
        if fermi is not None:
            ax1.set_ylabel(r"$E-E_f$ [eV]")
        else:
            ax1.set_ylabel(r"Energy [eV]")
        if elimit:
            ax1.set_ylim(elimit)

    elif bands_mode == "parametric":
        if dos_mode == "parametric":
            plot_bar = False
        else:
            plot_bar = True
        _, ax1 = plot.parametricPlot(
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            ticks=ticks,
            discontinuities=discontinuities,
            ax=ax1,
            plot_bar=plot_bar,
            mask=mask,
        )

        if fermi is not None:
            ax1.set_ylabel(r"$E-E_f$ [eV]")
        else:
            ax1.set_ylabel(r"Energy [eV]")
        if elimit is not None:
            ax1.set_ylim(elimit)

    if dos_mode == "plain":
        _, ax2 = dos_plot.plot_total(
            spins=dos_spins, ax=ax2, orientation="vertical", labels=dos_labels
        )
        dos = dos_plot.VaspXML.dos_total

    elif dos_mode == "parametric_line":
        if not total:
            _, ax2 = dos_plot.plot_parametric_line(
                atoms=atoms,
                spins=dos_spins,
                orbitals=orbitals,
                spin_colors=dos_spin_colors,
                ax=ax2,
                orientation="vertical",
                labels=dos_labels,
            )
            dos = dos_plot.VaspXML.dos_parametric(
                atoms=atoms, spin=dos_spins, orbitals=orbitals,
            )
        else:
            _, ax2 = dos_plot.plot_total(
                spins=dos_spins,
                spin_colors=[(0, 0, 0), (0, 0, 0)],
                ax=ax2,
                orientation="vertical",
            )
            dos = dos_plot.VaspXML.dos_total
            _, ax2 = dos_plot.plot_parametric_line(
                atoms=atoms,
                spins=dos_spins,
                orbitals=orbitals,
                spin_colors=dos_spin_colors,
                ax=ax2,
                orientation="vertical",
                labels=dos_labels,
            )
    elif dos_mode == "parametric":
        if not total:
            _, ax2 = dos_plot.plot_parametric(
                atoms=atoms,
                spins=dos_spins,
                orbitals=orbitals,
                spin_colors=dos_spin_colors,
                cmap=cmap,
                elimit=elimit,
                ax=ax2,
                vmin=vmin,
                vmax=vmax,
                orientation="vertical",
                labels=dos_labels,
            )
            dos = dos_plot.VaspXML.dos_total

        else:
            _, ax1 = dos_plot.plot_parametric(
                atoms=atoms,
                spins=dos_spins,
                orbitals=orbitals,
                spin_colors=dos_spin_colors,
                cmap=cmap,
                elimit=elimit,
                ax=ax2,
                vmin=vmin,
                vmax=vmax,
                orientation="vertical",
                labels=dos_labels,
            )
            dos = dos_plot.VaspXML.dos_total
            _, ax2 = dos_plot.plot_total(
                spins=dos_spins,
                spin_colors=[(0, 0, 0), (0, 0, 0)],
                ax=ax2,
                orientation="vertical",
            )
        ax2.yaxis.set_visible(False)
    elif dos_mode == "stack_species":
        if not total:
            _, ax2 = dos_plot.plot_stack_species(
                spins=dos_spins,
                spin_colors=dos_spin_colors,
                colors=dos_colors,
                elimit=elimit,
                figsize=(12, 6),
                ax=ax2,
                orientation="vertical",
            )
            dos = dos_plot.VaspXML.dos_total
        else:
            _, ax2 = dos_plot.plot_stack_species(
                spins=dos_spins,
                markersize=markersize,
                marker=marker,
                spin_colors=dos_spin_colors,
                colors=dos_colors,
                elimit=elimit,
                figsize=(12, 6),
                ax=ax2,
                orientation="vertical",
            )
            dos = dos_plot.VaspXML.dos_total
            _, ax2 = dos_plot.plot_total(
                spins=dos_spins,
                spin_colors=[(0, 0, 0), (0, 0, 0)],
                ax=ax2,
                orientation="vertical",
            )
    elif dos_mode == "stack_orbitals":
        if not total:
            _, ax2 = dos_plot.plot_stack_orbitals(
                spins=dos_spins,
                spin_colors=dos_spin_colors,
                colors=dos_colors,
                elimit=elimit,
                figsize=(12, 6),
                ax=ax2,
                orientation="vertical",
            )
            dos = dos_plot.VaspXML.dos_total
        else:
            _, ax2 = dos_plot.plot_stack_orbitals(
                spins=dos_spins,
                spin_colors=dos_spin_colors,
                colors=dos_colors,
                elimit=elimit,
                figsize=(12, 6),
                ax=ax2,
                orientation="vertical",
            )
            dos = dos_plot.VaspXML.dos_total
            _, ax2 = dos_plot.plot_total(
                spins=dos_spins,
                spin_colors=[(0, 0, 0), (0, 0, 0)],
                ax=ax2,
                orientation="vertical",
            )
    elif dos_mode == "stack":
        if not total:
            _, ax2 = dos_plot.plot_stack(
                items=dos_items,
                spins=dos_spins,
                spin_colors=dos_spin_colors,
                colors=dos_colors,
                elimit=elimit,
                figsize=(12, 6),
                ax=ax2,
                orientation="vertical",
            )

            dos = dos_plot.VaspXML.dos_total
        else:
            _, ax2 = dos_plot.plot_stack(
                items=dos_items,
                spins=dos_spins,
                spin_colors=dos_spin_colors,
                colors=dos_colors,
                elimit=elimit,
                figsize=(12, 6),
                ax=ax2,
                orientation="vertical",
            )
            dos = dos_plot.VaspXML.dos_total
            _, ax2 = dos_plot.plot_total(
                spins=dos_spins,
                spin_colors=[(0, 0, 0), (0, 0, 0)],
                ax=ax2,
                orientation="vertical",
            )

    ax2.set_ylim(elimit)
    ax2.axhline(color="black", linestyle="--")
    ax2.axvline(color="black", linestyle="--")
    if grid:
        ax1.grid()
        ax2.grid()
    #    ax2.yaxis.set_ticklabels([])

    if dos_labels or "stack" in dos_mode:
        ax2.legend()
    ax2.yaxis.set_visible(False)
    if (
        dos_mode == "stack_species"
        or dos_mode == "stack_orbitals"
        or dos_mode == "stack"
    ):
        ax2.set_xlim(ax2.get_xlim()[0], dos.values.max() * 1.4)

    if title:
        ax1.set_title(title)

    fig.tight_layout()

    if savefig:
        fig.savefig(savefig, bbox_inches="tight")
        plt.close()
        return None, None
    else:
        plt.show()
    return fig, ax1, ax2



"""
Phonon unfolding: Reciprocal space method. The method is described in
P. B. Allen et al. Phys Rev B 87, 085322 (2013).
This method should be also applicable to other bloch waves on discrete grid, eg. electrons wave function in wannier basis set, magnons, etc. Now only phonon istested.
"""
from ase.build import make_supercell
from ase.atoms import Atoms
import numpy as np


class Unfolder:
    """ phonon unfolding class"""

    def __init__(
        self,
        cell,
        basis,
        positions,
        supercell_matrix,
        eigenvectors,
        qpoints,
        tol_r=0.1,
        compare=None,
        phase=True,
    ):
        """
        Params:
        ===================
        cell: cell matrix. [a,b,c]
        basis: name of the basis. It's used to decide if two basis can be identical by translation. eg. for phonons, the basis can be ['x','y','z']*natoms, for electrons, it can be ['Ni|dxy','Mn|dxy'] if the two dxy are seen as different, or ['dxy','dxy'] if they are seen as the same.
        positions: positions(->basis).
        supercell matrix: The matrix that convert the primitive cell to supercell.
        eigenvectors: The phonon eigenvectors. format np.array() index=[ikpts, ifreq, 3*iatoms+j]. j=0..2
        qpoints: list of q-points.
        tol_r: tolerance. If abs(a-b) <r, they are seen as the same atom.
        """
        self._cell = cell
        self._basis = basis
        self._positions = positions
        self._scmat = supercell_matrix
        self._evecs = eigenvectors
        self._qpts = qpoints
        self._tol_r = tol_r
        self._trans_rs = None
        self._trans_indices = None
        self._make_translate_maps()
        self._phase = phase
        return

    def _translate(self, evec, r):
        """
        T(r) psi: r is integer numbers of primitive cell lattice matrix.
        Params:
        =================
        evec: an eigen vector of supercell
        r: The translate vector
        
        Returns:
        ================
         tevec: translated vector.
        """
        pass

    def _make_translate_maps(self):
        """
        find the mapping between supercell and translated cell.
        Returns:
        ===============
        A N * nbasis array.
        index[i] is the mapping from supercell to translated supercell so that
        T(r_i) psi = psi[indices[i]].
        
        TODO: vacancies/add_atoms not supported. How to do it? For vacancies, a ghost atom can be added. For add_atom, maybe we can just ignore them? Will it change the energy spectrum?
        """
        a1 = Atoms(symbols="H", positions=[(0, 0, 0)], cell=[1, 1, 1])
        sc = make_supercell(a1, self._scmat)
        rs = sc.get_scaled_positions()

        positions = self._positions
        indices = np.zeros([len(rs), len(positions)], dtype="int32")
        for i, ri in enumerate(rs):
            inds = []
            Tpositions = positions + np.array(ri)
            close_to_int = lambda x: np.all(np.abs(x - np.round(x)) < self._tol_r)
            for i_basis, pos in enumerate(positions):
                for j_basis, Tpos in enumerate(Tpositions):
                    dpos = Tpos - pos
                    if close_to_int(dpos) and (
                        self._basis[i_basis] == self._basis[j_basis]
                    ):
                        # indices[i, j_atom * self._ndim:j_atom * self._ndim + self._ndim] = np.arange(i_atom * self._ndim, i_atom * self._ndim + self._ndim)
                        indices[i, j_basis] = i_basis

        self._trans_rs = rs
        self._trans_indices = indices
        # print(indices)

    def get_weight(self, evec, qpt, G=None):
        """
        get the weight of a mode which has the wave vector of qpt and eigenvector of evec.
        W= sum_1^N < evec| T(r_i)exp(-I (K+G) * r_i| evec>, here G=0. T(r_i)exp(-I K r_i)| evec> = evec[indices[i]]
        """
        if G is None:
            G = np.zeros_like(qpt)
        weight = 0j
        N = len(self._trans_rs)
        for r_i, ind in zip(self._trans_rs, self._trans_indices):
            if self._phase:
                weight += (
                    np.vdot(evec, evec[ind])
                    * np.exp(1j * 2 * np.pi * np.dot(qpt + G, r_i))
                    / N
                )
            else:
                weight += (
                    np.vdot(evec, evec[ind])
                    / N
                    * np.exp(-1j * 2 * np.pi * np.dot(G, r_i))
                )
        return weight.real

    def get_weights(self):
        """
        Get the weight for all the modes.
        """
        nqpts, nfreqs = self._evecs.shape[0], self._evecs.shape[1]
        weights = np.zeros([nqpts, nfreqs])
        for iqpt in range(nqpts):
            for ifreq in range(nfreqs):
                weights[iqpt, ifreq] = self.get_weight(
                    self._evecs[iqpt, ifreq, :], self._qpts[iqpt]
                )

        self._weights = weights
        return self._weights



"""
Created on Fri March 2 2020
@author: Pedram Tavadze

"""

from re import findall

import numpy as np


class ElkParser:
    def __init__(self, elkin="elk.in", kdirect=True):

        # elk specific inp
        self.fin = elkin
        self.file_names = []
        self.elkin = None
        self.nbands = None

        self.nspin = None
        self.kticks = None

        self.kdirect = kdirect

        self.kpoints = None
        self.bands = None

        self.spd = None
        self.cspd = None

        # spin polarized parameters
        self.spinpol = None

        self.orbitalName = [
            "Y00",
            "Y1-1",
            "Y10",
            "Y11",
            "Y2-2",
            "Y2-1",
            "Y20",
            "Y21",
            "Y22",
            "Y3-3",
            "Y3-2",
            "Y3-1",
            "Y30",
            "Y3-1",
            "Y3-2",
            "Y3-3",
        ]
        self.orbital_name_short = ["s", "p", "d", "f"]
        self.norbital = None  # number of orbitals

        # number of spin components (blocks of data), 1: non-magnetic non
        # polarized, 2: spin polarized collinear, 4: non-collinear
        # spin.
        # NOTE: before calling to `self._readOrbital` the case '4'
        # is marked as '1'

        self._read_elkin()
        self._read_bands()

        return

    #    @property
    #    def nspin(self):
    #        """
    #        number of spin, default is 1.
    #        """
    #        nspindict = {1: 1, 2: 2, 4: 2, None: 1}
    #        return nspindict[self.ispin]

    @property
    def spd_orb(self):
        # indices: ikpt, iband, ispin, iion, iorb
        # remove indices and total from iorb.
        return self.spd[:, :, :, 1:-1]

    @property
    def nhigh_sym(self):
        """
        Returns nu,ber of high symmetry points
        """
        return int(findall("plot1d\n\s*([0-9]*)\s*[0-9]*", self.elkin)[0])

    @property
    def nkpoints(self):
        """
        Returns total number of kpoints
        """
        return int(findall("plot1d\n\s*[0-9]*\s*([0-9]*)", self.elkin)[0])

    @property
    def tasks(self):
        """
        Returns the tasks calculated by elk
        """
        return [
            int(x) for x in findall("tasks\n\s*([0-9\s\n]*)", self.elkin)[0].split()
        ]

    @property
    def high_symmetry_points(self):
        """
        Returns the corrdinates of high symmtery points provided in elk.in
        """
        if 20 in self.tasks or 21 in self.tasks or 22 in self.tasks:
            raw_hsp = findall(
                "plot1d\n\s*[0-9]*\s*[0-9]*.*\n" + self.nhigh_sym * "([0-9\s\.-]*).*\n",
                self.elkin,
            )[0]
            high_symmetry_points = np.zeros(shape=(self.nhigh_sym, 3))
            for ihs in range(self.nhigh_sym):
                high_symmetry_points[ihs, :] = [
                    float(x) for x in raw_hsp[ihs].split()[0:3]
                ]
            return high_symmetry_points

    @property
    def nspecies(self):
        """
        Returns number of species
        """

        return int(findall("atoms\n\s*([0-9]*)", self.elkin)[0])

    @property
    def natom(self):
        """
        Returns number of atoms in the structure
        """
        natom = 0
        for ispc in findall("'([A-Za-z]*).in'.*\n\s*([0-9]*)", self.elkin):
            natom += int(ispc[1])
        return natom

    @property
    def composition(self):
        """
        Returns the composition of the structure
        """
        composition = {}
        for ispc in findall("'([A-Za-z]*).in'.*\n\s*([0-9]*)", self.elkin):
            composition[ispc[0]] = int(ispc[1])
        return composition

    @property
    def knames(self):
        """
        Returns the names of the high symmetry points(x ticks) provided in elk.in

        """
        raw_ticks = findall(
            "plot1d\n\s*[0-9]*\s*[0-9]*.*\n" + self.nhigh_sym * ".*:(.*)\n", self.elkin
        )[0]
        if len(raw_ticks) != self.nhigh_sym:
            knames = [str(x) for x in range(self.nhigh_sym)]
        else:
            knames = [
                "$%s$" % (x.replace(",", "").replace("vlvp1d", "").replace(" ", ""))
                for x in findall(
                    "plot1d\n\s*[0-9]*\s*[0-9]*.*\n" + self.nhigh_sym * ".*:(.*)\n",
                    self.elkin,
                )[0]
            ]
        return knames

    def _read_elkin(self):
        """
        Reads and parses elk.in
        """
        rf = open(self.fin, "r")
        self.elkin = rf.read()
        rf.close()

        if 20 in self.tasks:
            self.file_names.append("BANDS.OUT")
        if 21 in self.tasks or 22 in self.tasks:
            ispc = 1
            for spc in self.composition:
                for iatom in range(self.composition[spc]):
                    self.file_names.append(
                        "BAND_S{:02d}_A{:04d}.OUT".format(ispc, iatom + 1)
                    )
                ispc += 1

        # Checking if spinpol = .true. in elk.in
        self.spinpol = findall(r"spinpol\s*([.a-zA-Z]*)", self.elkin)[0]
        if self.spinpol:
            if self.spinpol == ".true.":
                print("\nElk colinear spin calculation detected.\n")
                self.spinpol = True
                self.nspin = 2
            else:
                print("\nElk non spin calculation detected.\n")
                self.nspin = 1
        else:
            print(
                "\nNo spinpol keyword found in elk.in. Assuming non spin calculation.\n"
            )
            self.nspin = 1

    def _read_bands(self):
        """
        if the task is any of 20,21,22 it parses the BANDS*.OUT files
        and prepares the spd, bands and kpoints for bandsplot
        """
        if np.any([x in self.tasks for x in [20, 21, 22]]):

            rf = open(self.file_names[0], "r")
            lines = rf.readlines()
            rf.close()

            self.nbands = int(len(lines) / (self.nkpoints + 1))
            self.bands = np.zeros(shape=(self.nkpoints, self.nbands))

            rf = open("BANDLINES.OUT", "r")
            bandLines = rf.readlines()
            rf.close()

            tick_pos = []
            # using strings for a better comparision and avoiding rounding by python
            for iline in range(0, len(bandLines), 3):
                tick_pos.append(bandLines[iline].split()[0])
            x_points = []
            for iline in range(self.nkpoints):
                x_points.append(lines[iline].split()[0])

            self.kpoints = np.zeros(shape=(self.nkpoints, 3))
            x_points = np.array(x_points)
            tick_pos = np.array(tick_pos)

            self.kticks = []
            for ihs in range(1, self.nhigh_sym):
                start = np.where(x_points == tick_pos[ihs - 1])[0][0]
                end = np.where(x_points == tick_pos[ihs])[0][0] + 1
                self.kpoints[start:end][:] = np.linspace(
                    self.high_symmetry_points[ihs - 1],
                    self.high_symmetry_points[ihs],
                    end - start,
                )

                self.kticks.append(start)
            self.kticks.append(self.nkpoints - 1)
            if not self.kdirect:
                self.kpoints = np.dot(self.kpoints, self.reclat)

            rf = open(self.file_names[0], "r")
            lines = rf.readlines()
            rf.close()

            iline = 0
            for iband in range(self.nbands):
                for ikpoint in range(self.nkpoints):
                    self.bands[ikpoint, iband] = float(lines[iline].split()[1])
                    iline += 1
                if ikpoint == self.nkpoints - 1:
                    iline += 1
            self.bands *= 27.21138386

            self.norbital = 16
            self.spd = np.zeros(
                shape=(
                    self.nkpoints,
                    self.nbands,
                    self.nspin,
                    self.natom + 1,
                    self.norbital + 2,
                )
            )
            idx_bands_out = None
            for ifile in range(len(self.file_names)):
                if self.file_names[ifile] == "BANDS.OUT":
                    idx_bands_out = ifile
            if idx_bands_out != None:
                del self.file_names[idx_bands_out]

            for ifile in range(self.natom):
                rf = open(self.file_names[ifile], "r")
                lines = rf.readlines()
                rf.close()
                iline = 0

                for iband in range(self.nbands):
                    for ikpoint in range(self.nkpoints):
                        temp = np.array([float(x) for x in lines[iline].split()])
                        self.spd[ikpoint, iband, 0, ifile, 0] = ifile + 1
                        self.spd[ikpoint, iband, 0, ifile, 1:-1] = temp[2:]
                        iline += 1
                    if ikpoint == self.nkpoints - 1:
                        iline += 1
            # self.spd[:,:,:,-1,:] = self.spd.sum(axis=3)
            self.spd[:, :, :, :, -1] = np.sum(self.spd[:, :, :, :, 1:-1], axis=4)
            self.spd[:, :, :, -1, :] = self.spd.sum(axis=3)
            self.spd[:, :, 0, -1, 0] = 0

            if self.nspin == 2:
                # spin up block for spin = 1
                self.spd[:, : self.nbands // 2, 1, :, :] = self.spd[
                    :, : self.nbands // 2, 0, :, :
                ]
                # spin down block for spin = 1
                self.spd[:, self.nbands // 2 :, 1, :, :] = (
                    -1 * self.spd[:, self.nbands // 2 :, 0, :, :]
                )

                # manipulating spd array for spin polarized calculations.
                # The shape is (nkpoints,2*nbands,2,natoms,norbitals)
                # The third dimension is for spin.
                # When this is zero, the bands*2 (all spin up and down bands) have positive projections.
                # When this is one, the the first half of bands (spin up) will have positive projections
                # and the second half (spin down) will have negative projections. This is to adhere to
                # the convention used in PyProcar to obtain spin density and spin magnetization.

                # spin up and spin down block for spin = 0
                # spd2[:, :, 0, :, :] = self.spd[:, :, 0, :, :]

    @property
    def fermi(self):
        """
        Returns the fermi energy read from FERMI.OUT
        """
        rf = open("EFERMI.OUT", "r")
        fermi = float(rf.readline().split()[0])
        rf.close()
        return fermi

    @property
    def reclat(self):
        """
        Returns the reciprocal lattice read from LATTICE.OUT
        """
        rf = open("LATTICE.OUT", "r")
        data = rf.read()
        rf.close()
        lattice_block = findall(r"matrix\s*:([\s0-9.]*)Inverse", data)
        lattice_array = np.array(lattice_block[1].split(), dtype=float)
        reclat = lattice_array.reshape((3, 3))
        return reclat



import numpy as np
import re
import logging
import matplotlib.pyplot as plt
import sys
from ..utilsprocar import UtilsProcar


class ProcarParser:
    """Parses a PROCAR file and store it in memory. It only deals with
    PROCAR files, that means no Fermi energy (UtilsProcar.FermiOutcar
    can help), and the reciprocal vectors should be supplied (if used,
    see UtilsProcar class).

    Members:

    __init__(self, loglevel): The setup the variables internally, `loglevel`
      sets the verbosity level ie: `loglevel=logging.DEBUG` for debugging. Its
      default is `logging.WARNING`

    readFile(self, procar=None, permissive=False, recLattice=None):
      The only method of the API it load the file completely.

      Arguments:
    `procar=None`: name of the PROCAR file, can be a gzipped file (the
                    extension is no required). The default covers a wide range
                    of obvious alternatives.
      `permissive=False`: Set to `True` if the PROCAR file has problems reading
                        the Kpoints (stupid Fortran), but in that case the
                          Kpoints mesh will be discarded. Future updates could
                          allow it to handle other formating/corruption issues.
      `recLattice`=None: Reciprical Vectors, you want to provide them since not
                        all the paths on the BZ are the same.

    Don't use the other methods beggining with underscores "_"

    Example:
    To read a PROCAR or PROCAR.gz file:
    >>> foo = ProcarParser()
    >>> foo.readFile()

    To include the reciprocal vectors, and file name MyFirstPROCAR
    >>> outcarparser = UtilsProcar()
    >>> recLat = outcarparser.RecLatOutcar(args.outcar)
    >>> foo = ProcarParser()
    >>> foo.readFile("MyFirstPROCAR", recLat=recLat)

    """

    def __init__(self, loglevel=logging.WARNING):
        # array with k-points, they have the following values
        # -None: if not parsed (yet) or parsed with a `permissive` flag on
        # -direct coordinates: if a recLattice was not supplied to the parser
        # -cartesian coords: if a recLattice was supplied to the parser.
        # In the later cases, self.kpoints.shape=(self.kpointsCount, 3)
        self.kpoints = None
        # Number of kpoints, as given by the KPOINTS header (PROCAR file)
        self.kpointsCount = None

        # bands headers present in PROCAR file.
        # self.bands.shape=(self.kpointsCount,self.bandsCount)
        self.bands = None
        # Number of bands. For a spin polarized calculation the number of
        # bands is double (spin ip + spin down). On this array there is no
        # distinction between spin up and down
        self.bandsCount = None

        # Number of ions+1 the +1 is the 'tot' field, ie: the sum over all atoms
        self.ionsCount = None

        self.fileStr = None  # the actual file, stored in memory
        self.spd = None  # the atom/orbital projected data
        self.cspd = None  # spd data with the phase
        self.orbitalName = [
            "s",
            "py",
            "pz",
            "px",
            "dxy",
            "dyz",
            "dz2",
            "dxz",
            "x2-y2",
            "fy3x2",
            "fxyz",
            "fyz2",
            "fz3",
            "fxz2",
            "fzx2",
            "fx3",
            "tot",
        ]
        self.orbitalName_old = [
            "s",
            "py",
            "pz",
            "px",
            "dxy",
            "dyz",
            "dz2",
            "dxz",
            "dx2",
            "tot",
        ]
        self.orbitalName_short = ["s", "p", "d", "f", "tot"]
        self.orbitalCount = None  # number of orbitals

        # number of spin components (blocks of data), 1: non-magnetic non
        # polarized, 2: spin polarized collinear, 4: non-collinear
        # spin.
        # NOTE: before calling to `self._readOrbital` the case '4'
        # is marked as '1'
        self.ispin = None
        self.recLattice = None  # reciprocal lattice vectors
        self.utils = UtilsProcar()

        self.log = logging.getLogger("ProcarParser")
        self.log.setLevel(loglevel)
        self.ch = logging.StreamHandler()
        self.ch.setFormatter(
            logging.Formatter("%(name)s::%(levelname)s:" " %(message)s")
        )
        self.ch.setLevel(logging.DEBUG)
        self.log.addHandler(self.ch)
        # At last, one message to the logger.
        self.log.debug("Procar instanciated")
        return

    @property
    def nspin(self):
        """
        number of spin, default is 1.
        """
        nspindict = {1: 1, 2: 2, 4: 2, None: 1}
        return nspindict[self.ispin]

    @property
    def spd_orb(self):
        # indices: ikpt, iband, ispin, iion, iorb
        # remove indices and total from iorb.
        return self.spd[:, :, :, 1:-1]

    def _readKpoints(self, permissive=False):
        """Reads the k-point headers. A typical k-point line is:
        k-point    1 :    0.00000000 0.00000000 0.00000000  weight = 0.00003704\n
        fills self.kpoint[kpointsCount][3]
        The weights are discarded (are they useful?)
        """
        self.log.debug("readKpoints")
        if not self.fileStr:
            log.warning("You should invoke `procar.readFile()` instead. Returning")
            return

        # finding all the K-points headers
        self.kpoints = re.findall(r"k-point\s+\d+\s*:\s+([-.\d\s]+)", self.fileStr)
        self.log.debug(str(len(self.kpoints)) + " K-point headers found")
        self.log.debug("The first match found is: " + str(self.kpoints[0]))

        # trying to build an array
        self.kpoints = [x.split() for x in self.kpoints]
        try:
            self.kpoints = np.array(self.kpoints, dtype=float)
        except ValueError:
            self.log.error("Ill-formatted data:")
            print("\n".join([str(x) for x in self.kpoints]))
            if permissive is True:
                # Discarding the kpoints list, however I need to set
                # self.ispin beforehand.
                if len(self.kpoints) == self.kpointsCount:
                    self.ispin = 1
                elif len(self.kpoints) == 2 * self.kpointsCount:
                    self.ispin = 2
                else:
                    raise ValueError("Kpoints do not match with ispin=1 or 2.")
                self.kpoints = None
                self.log.warning("K-points list is useless, setting it to `None`")
                return
            else:
                raise ValueError("Badly formated Kpoints headers, try `--permissive`")
        # if successful, go on

        # trying to identify an non-polarized or non-collinear case, a
        # polarized case or a defective file

        if len(self.kpoints) != self.kpointsCount:
            # if they do not match, may means two things a spin polarized
            # case or a bad file, lets check
            self.log.debug(
                "Number of kpoints do not match, looking for a " "spin-polarized case"
            )
            # lets start testing if it is spin polarized, if so, there
            # should be 2 identical blocks of kpoints.
            up, down = np.vsplit(self.kpoints, 2)
            if (up == down).all():
                self.log.info("Spin-polarized calculation found")
                self.ispin = 2
                # just keeping one set of kpoints (the other will be
                # discarded)
                self.kpoints = up
            else:
                self.log.error("Number of K-points do not match! check them.")
                raise RuntimeError("Bad Kpoints list.")
        # if ISPIN != 2 setting ISPIN=1 (later for the non-collinear case 1->4)
        # It is unknown until parsing the projected data
        else:
            self.ispin = 1

        # checking again, for compatibility,
        if len(self.kpoints) != self.kpointsCount:
            raise RuntimeError(
                "Kpoints number do not match with metadata (header of PROCAR)"
            )

        self.log.debug(str(self.kpoints))
        self.log.info("The kpoints shape is " + str(self.kpoints.shape))

        if self.recLattice is not None:
            self.log.info("Changing to cartesians coordinates")
            self.kpoints = np.dot(self.kpoints, self.recLattice)
            self.log.debug("New kpoints: \n" + str(self.kpoints))
        return

    def _readBands(self):
        """Reads the bands header. A typical bands is:
        band   1 # energy   -7.11986315 # occ.  1.00000000

        fills self.bands[kpointsCount][bandsCount]

        The occupation numbers are discarded (are they useful?)"""
        self.log.debug("readBands")
        if not self.fileStr:
            log.warning("You should invoke `procar.read()` instead. Returning")
            return

        # finding all bands
        self.bands = re.findall(
            r"band\s*(\d+)\s*#\s*energy\s*([-.\d\s]+)", self.fileStr
        )
        self.log.debug(
            str(len(self.bands))
            + " bands headers found, bands*Kpoints = "
            + str(self.bandsCount * self.kpointsCount)
        )
        self.log.debug("The first match found is: " + str(self.bands[0]))

        # checking if the number of bands match

        if len(self.bands) != self.bandsCount * self.kpointsCount * self.ispin:
            self.log.error("Number of bands headers do not match")
            raise RuntimeError("Number of bands don't match")

        # casting to array to manipulate the bands
        self.bands = np.array(self.bands, dtype=float)
        self.log.debug(str(self.bands))

        # Now I will deal with the spin polarized case. The goal is join
        # them like for a non-magnetic case
        if self.ispin == 2:
            # up and down are along the first axis
            up, down = np.vsplit(self.bands, 2)
            self.log.debug("up   , " + str(up.shape))
            self.log.debug("down , " + str(down.shape))

            # reshapping (the 2  means both band index and energy)
            up.shape = (self.kpointsCount, self.bandsCount, 2)
            down.shape = (self.kpointsCount, self.bandsCount, 2)

            # setting the correct number of bands (up+down)
            self.bandsCount *= 2
            self.log.debug("New number of bands : " + str(self.bandsCount))

            # and joining along the second axis (axis=1), ie: bands-like
            self.bands = np.concatenate((up, down), axis=1)

        # otherwise just reshaping is needed
        else:
            self.bands.shape = (self.kpointsCount, self.bandsCount, 2)

        # Making a test if the broadcast is rigth, otherwise just print
        test = [x.max() - x.min() for x in self.bands[:, :, 0].transpose()]
        if np.array(test).any():
            self.log.warning(
                "The indexes of bands do not match. CHECK IT. "
                "Likely the data was wrongly broadcasted"
            )
            self.log.warning(str(self.bands[:, :, 0]))
        # Now safely removing the band index
        self.bands = self.bands[:, :, 1]
        self.log.info("The bands shape is " + str(self.bands.shape))
        return

    def _readOrbital(self):
        """Reads all the spd-projected data. A typical/expected block is:
    ion      s     py     pz     px    dxy    dyz    dz2    dxz    dx2    tot
      1  0.079  0.000  0.001  0.000  0.000  0.000  0.000  0.000  0.000  0.079
      2  0.152  0.000  0.000  0.000  0.000  0.000  0.000  0.000  0.000  0.152
      3  0.079  0.000  0.001  0.000  0.000  0.000  0.000  0.000  0.000  0.079
      4  0.188  0.000  0.000  0.000  0.000  0.000  0.000  0.000  0.000  0.188
      5  0.188  0.000  0.000  0.000  0.000  0.000  0.000  0.000  0.000  0.188
    tot  0.686  0.000  0.002  0.000  0.000  0.000  0.000  0.000  0.000  0.688
    (x2 for spin-polarized -akwardkly formatted-, x4 non-collinear -nicely
    formatted-).

    The data is stored in an array self.spd[kpoint][band][ispin][atom][orbital]

    Undefined behavior in case of phase factors (LORBIT = 12).
    """
        self.log.debug("readOrbital")
        if not self.fileStr:
            log.warning("You should invoke `procar.readFile()` instead. Returning")
            return

        # finding all orbital headers
        self.spd = re.findall(r"ion(.+)", self.fileStr)
        self.log.info("the first orbital match reads: " + self.spd[0])
        self.log.debug("And I found " + str(len(self.spd)) + " orbitals headers")

        # testing if the orbital names are known (the standard ones)
        FoundOrbs = self.spd[0].split()
        size = len(FoundOrbs)
        # only the first 'size' orbital
        StdOrbs = self.orbitalName[: size - 1] + self.orbitalName[-1:]
        StdOrbs_short = self.orbitalName_short[: size - 1] + self.orbitalName_short[-1:]
        StdOrbs_old = self.orbitalName_old[: size - 1] + self.orbitalName_old[-1:]
        if (
            FoundOrbs != (StdOrbs)
            and FoundOrbs != (StdOrbs_short)
            and FoundOrbs != (StdOrbs_old)
        ):
            self.log.warning(
                str(size) + " orbitals. (Some of) They are unknow (if "
                "you did 'filter' them it is OK)."
            )
        self.orbitalCount = size
        self.orbitalNames = self.spd[0].split()
        self.log.debug(
            "Anyway, I will use the following set of orbitals: "
            + str(self.orbitalNames)
        )

        # Now reading the bulk of data
        self.log.debug("Now searching the values")
        # The case of just one atom is handled differently since the VASP
        # output is a little different
        if self.ionsCount == 1:
            self.spd = re.findall(r"^(\s*1\s+.+)$", self.fileStr, re.MULTILINE)
        else:
            # Added by Francisco to speed up filtering on June 4th, 2019
            # get rid of phase factors
            self.spd = re.findall(r"ion.+tot\n([-.\d\seto]+)", self.fileStr)
            self.spd = "".join(self.spd)
            self.spd = re.findall(r"([-.\d\se]+tot.+)\n", self.spd)
        # free the memory (could be a lot)
        self.fileStr = None
        self.log.debug("the first entry is \n" + self.spd[0])

        # Now the method will try to find the value of self.ispin,
        # previously it was set to either 1 or 2. If "1", it could be 1 or
        # 4, but previously it was impossible to find the rigth value. If
        # "2" it has to macth with the number of entries of spd data.

        self.log.debug("Number of entries found: " + str(len(self.spd)))
        expected = self.bandsCount * self.kpointsCount
        self.log.debug(
            "The number of entries for a non magnetic calc. is: " + str(expected)
        )
        if expected == len(self.spd):
            self.log.info("Both numbers match, ok, going ahead")
        # catching a non-collinear calc.
        elif expected * 4 == len(self.spd):
            self.log.info("non-collinear calculation found")
            # testing if previous ispin value is ok
            if self.ispin != 1:
                self.log.warning(
                    "Incompatible data: self.ispin= " + str(self.ispin) + ". Now is 4"
                )
            self.ispin = 4
        else:
            self.log.error("The parser or data is wrong!!!")
            self.log.info("bandsCount: " + str(self.bandsCount))
            self.log.info("KpointsCount: " + str(self.kpointsCount))
            raise RuntimeError("Shit happens")

        # checking for consistency
        for line in self.spd:
            if len(line.split()) != (self.ionsCount) * (self.orbitalCount + 1):
                self.log.error(
                    "Expected: "
                    + str(self.ionsCount)
                    + "*"
                    + str(self.orbitalCount + 1)
                    + " = "
                    + str((self.ionsCount) * (self.orbitalCount + 1))
                    + " Fields. Present block: "
                    + str(len(line.split()))
                )
                raise RuntimeError("Flats happens")

        # replacing the "tot" string by a number, to allows a conversion
        # to numpy
        self.spd = [x.replace("tot", "0") for x in self.spd]
        self.spd = [x.split() for x in self.spd]
        self.spd = np.array(self.spd, dtype=float)
        self.log.debug("The spd (old) array shape is:" + str(self.spd.shape))

        # handling collinear polarized case
        if self.ispin == 2:
            self.log.debug("Handling spin-polarized collinear case...")
            # splitting both spin components, now they are along k-points
            # axis (1st axis) but, then should be concatenated along the
            # bands.
            up, down = np.vsplit(self.spd, 2)
            # ispin = 1 for a while, we will made the distinction
            up.shape = (
                self.kpointsCount,
                int(self.bandsCount / 2),
                1,
                self.ionsCount,
                self.orbitalCount + 1,
            )
            down.shape = (
                self.kpointsCount,
                int(self.bandsCount / 2),
                1,
                self.ionsCount,
                self.orbitalCount + 1,
            )
            # concatenating bandwise. Density and magntization, their
            # meaning is obvious, and do uses 2 times more memory than
            # required, but I *WANT* to keep it as close as possible to the
            # non-collinear or non-polarized case
            density = np.concatenate((up, down), axis=1)
            magnet = np.concatenate((up, -down), axis=1)
            # concatenated along 'ispin axis'
            self.spd = np.concatenate((density, magnet), axis=2)
            self.log.debug("polarized collinear spd.shape= " + str(self.spd.shape))

        # otherwise, just a reshaping suffices
        else:
            self.spd.shape = (
                self.kpointsCount,
                self.bandsCount,
                self.ispin,
                self.ionsCount,
                self.orbitalCount + 1,
            )

        self.log.info("spd array ready. Its shape is:" + str(self.spd.shape))
        return

    def readFile(self, procar=None, phase=False, permissive=False, recLattice=None):
        """Reads and parses the whole PROCAR file. This method is a sort
    of metamethod: it opens the file, reads the meta data and call the
    respective functions for parsing kpoints, bands, and projected
    data.

    Args:

    -procar: The file name, if `None` or a directory, a suitable set
     of defaults will be used. Default=None

    -permissive: turn on (or off) some features to deal with badly
     written PROCAR files (stupid fortran), up to now just ignores the
     kpoints coordinates, which -as side effect- prevent he rigth
     space between kpoints. Default=False (off)


    -recLattice: a 3x3 array containing the reciprocal vectors, to
     change the Kpoints from rec. coordinates to cartesians. Rarely
     given by hand, see `UtilsProcar.RecLatProcar`. If given, the
     kpoints will be converted from direct coordinates to cartesian
     ones. Default=None

    """
        self.log.debug("readFile...")

        self.recLattice = recLattice

        self.log.debug("Opening file: '" + str(procar) + "'")
        f = self.utils.OpenFile(procar)
        # Line 1: PROCAR lm decomposed
        f.readline()  # throwaway
        # Line 2: # of k-points:  816   # of bands:  52   # of ions:   8
        metaLine = f.readline()  # metadata
        self.log.debug("The metadata line is: " + metaLine)
        re.findall(r"#[^:]+:([^#]+)", metaLine)
        self.kpointsCount, self.bandsCount, self.ionsCount = map(
            int, re.findall(r"#[^:]+:([^#]+)", metaLine)
        )
        self.log.info("kpointsCount = " + str(self.kpointsCount))
        self.log.info("bandsCount = " + str(self.bandsCount))
        self.log.info("ionsCount = " + str(self.ionsCount))
        if self.ionsCount == 1:
            self.log.warning(
                "Special case: only one atom found. The program may not work as expected"
            )
        else:
            self.log.debug("An extra ion representing the  total value will be added")
            self.ionsCount = self.ionsCount + 1

        # reading all the rest of the file to be parsed below
        self.fileStr = f.read()
        self._readKpoints(permissive)
        self._readBands()
        self._readOrbital()
        self.log.debug("readfile...done")
        return

    # Slower way to parse the phase included PROCAR. He Xu implemented a faster way. See readFile2()
    # if phase == False:
    #     self.log.debug("readFile...")

    #     self.recLattice = recLattice

    #     self.log.debug("Opening file: '" + str(procar) + "'")
    #     f = self.utils.OpenFile(procar)
    #     # Line 1: PROCAR lm decomposed
    #     f.readline()  # throwaway
    #     # Line 2: # of k-points:  816   # of bands:  52   # of ions:   8
    #     metaLine = f.readline()  # metadata
    #     self.log.debug("The metadata line is: " + metaLine)
    #     re.findall(r"#[^:]+:([^#]+)", metaLine)
    #     self.kpointsCount, self.bandsCount, self.ionsCount = \
    #         list(map(int, re.findall(r"#[^:]+:([^#]+)", metaLine)))
    #     self.log.info("kpointsCount = " + str(self.kpointsCount))
    #     self.log.info("bandsCount = " + str(self.bandsCount))
    #     self.log.info("ionsCount = " + str(self.ionsCount))
    #     if self.ionsCount is 1:
    #         self.log.warning(
    #             "Special case: only one atom found. The program may not work as expected"
    #         )
    #     else:
    #         self.log.debug(
    #             "An extra ion representing the  total value will be added")
    #         self.ionsCount = self.ionsCount + 1

    #     #reading all the rest of the file to be parsed below
    #     self.fileStr = f.read()
    #     self._readKpoints(permissive)
    #     self._readBands()
    #     self._readOrbital()
    #     self.log.debug("readfile...done")

    # elif phase == True:  #for LORBIT=12
    #     self.recLattice = recLattice
    #     f = self.utils.OpenFile(procar)
    #     f.readline()  #throw away first line
    #     metaLine = f.readline()  # header

    #     #parsing header information
    #     self.kpointsCount, self.bandsCount, self.ionsCount = list(
    #         map(int, re.findall(r"#[^:]+:([^#]+)", metaLine)))

    #     if self.ionsCount is 1:
    #         self.log.warning(
    #             "Special case: only one atom found. The program may not work as expected"
    #         )
    #     else:
    #         self.log.debug(
    #             "An extra ion representing the  total value will be added")
    #         self.ionsCount = self.ionsCount + 1

    #     #reading all the rest of the file to be parsed below saved as spd
    #     self.fileStr = f.read()
    #     self._readKpoints(permissive)
    #     self._readBands()
    #     self._readOrbital()
    #     self.log.debug("readfile...done")

    #     #reading the complex phase data. will be saved as cspd
    #     f2 = self.utils.OpenFile(procar)
    #     f2.readline()  #throw away first line
    #     f2.readline()  #throw away header line
    #     data2 = f2.read()

    #     #parsing
    #     spd0 = re.findall(r"ion(.+)", data2)  #headers of the blocks
    #     FoundOrbs = spd0[1].split()
    #     size = len(FoundOrbs)
    #     corbitalCount = size
    #     spd_phase = re.findall(
    #         r"(?<=dx2-y2)([charge0-9.\s-]*)(?=band|k-point|\Z)",
    #         data2)  #for LORBIT=12

    #     spd_new = []
    #     for i in range(len(spd_phase)):
    #         #get last list of original block and replace spd last line and append all to get new spd.
    #         # for charge line use spd instead of spd_real
    #         spd_last = spd_phase[i].split()[-(corbitalCount + 2):]
    #         result = []
    #         for counter, value in enumerate(spd_last):
    #             result.append(value)
    #             result.append('0')
    #         del result[1]
    #         del result[-1]

    #         #replace last line of each spd block
    #         spd_block = spd_phase[i].split()
    #         spd_block[-(corbitalCount + 2):] = result
    #         spd_block = [x.replace('charge', '0') for x in spd_block]
    #         spd_new.append(spd_block)

    #     # conversion to numpy
    #     spd_phase = np.array(spd_new, dtype=float)

    #     #reshaping
    #     spd_phase.shape = (self.kpointsCount, self.bandsCount, 1,
    #                        self.ionsCount, 2 * corbitalCount + 2)

    #     #matrix to hold complex values
    #     self.cspd = np.zeros([
    #         self.kpointsCount, self.bandsCount, 1, self.ionsCount,
    #         corbitalCount + 2
    #     ],
    #                          dtype='complex')

    #     for ikpointsCount in range(self.kpointsCount):
    #         for ibandsCount in range(self.bandsCount):
    #             for iionsCount in range(self.ionsCount):
    #                 orbs_real = spd_phase[ikpointsCount][ibandsCount][0][
    #                     iionsCount][1:-1:2]
    #                 orbs_imag = spd_phase[ikpointsCount][ibandsCount][0][
    #                     iionsCount][2::2]
    #                 orbs_all = orbs_real + (1j * orbs_imag)
    #                 self.cspd[ikpointsCount, ibandsCount, 0,
    #                           iionsCount, :] = np.concatenate(
    #                               [[
    #                                   spd_phase[ikpointsCount][ibandsCount]
    #                                   [0][iionsCount][0]
    #                               ], orbs_all,
    #                                [
    #                                    spd_phase[ikpointsCount]
    #                                    [ibandsCount][0][iionsCount][-1]
    #                                ]])

    # return

    def readFile2(
        self,
        procar=None,
        phase=False,
        permissive=False,
        recLattice=None,
        ispin=None,  # the only spin channle to read
    ):
        """
        Read file in a line by line manner.
        Only used when the phase factor is in procar. (for vasp, lorbit=12)
        """
        # Fall back to readFile function if no phase
        self.bands = None
        if not phase:
            self.readFile(
                procar=procar, phase=False, permissive=permissive, recLattice=recLattice
            )
        else:
            if ispin is None:
                nspin = 1
            else:
                nspin = 2
            iispin = 0
            self.projections = None
            ikpt = 0
            iband = 0
            nkread = 0
            # with open(self.fname) as myfile:
            f = self.utils.OpenFile(procar)
            lines = iter(f.readlines())
            last_iband = -1
            for line in lines:
                if line.startswith("# of k-points"):
                    a = re.findall(":\s*([0-9]*)", line)
                    self.kpointsCount, self.bandsCount, self.ionsCount = map(int, a)
                    self.kpoints = np.zeros([self.kpointsCount, 3])
                    self.kweights = np.zeros(self.kpointsCount)
                    if self.bands is None:
                        self.bands = np.zeros(
                            [nspin, self.kpointsCount, self.bandsCount]
                        )
                if line.strip().startswith("k-point"):
                    ss = line.strip().split()
                    ikpt = int(ss[1]) - 1
                    k0 = float(ss[3])
                    k1 = float(ss[4])
                    k2 = float(ss[5])
                    w = float(ss[-1])
                    self.kpoints[ikpt, :] = [k0, k1, k2]
                    self.kweights[ikpt] = w
                    nkread += 1
                    if nkread <= self.kpointsCount:
                        iispin = 0
                    else:
                        iispin = 1
                if line.strip().startswith("band"):
                    ss = line.strip().split()
                    try:
                        iband = int(ss[1]) - 1
                    except ValueError:
                        iband = last_iband + 1
                    last_iband = iband
                    e = float(ss[4])
                    occ = float(ss[-1])
                    self.bands[iispin, ikpt, iband] = e
                if line.strip().startswith("ion"):
                    if line.strip().endswith("tot"):
                        self.orbitalName = line.strip().split()[1:-1]
                        self.orbitalCount = len(self.orbitalName)
                    if self.projections is None:
                        self.projections = np.zeros(
                            [
                                self.kpointsCount,
                                self.bandsCount,
                                self.ionsCount,
                                self.orbitalCount,
                            ]
                        )
                        self.carray = np.zeros(
                            [
                                self.kpointsCount,
                                self.bandsCount,
                                nspin,
                                self.ionsCount,
                                self.orbitalCount,
                            ],
                            dtype="complex",
                        )
                    for i in range(self.ionsCount):
                        line = next(lines)
                        t = line.strip().split()
                        if len(t) == self.orbitalCount + 2:
                            self.projections[ikpt, iband, iispin, :] = [
                                float(x) for x in t[1:-1]
                            ]
                        elif len(t) == self.orbitalCount * 2 + 2:
                            self.carray[ikpt, iband, iispin, i, :] += np.array(
                                [float(x) for x in t[1:-1:2]]
                            )
                            self.carray[ikpt, iband, iispin, i, :] += 1j * np.array(
                                [float(x) for x in t[2::2]]
                            )

                        # Added by Francisco to parse older version of PROCAR format on Jun 11, 2019
                        elif len(t) == self.orbitalCount * 1 + 1:
                            self.carray[ikpt, iband, iispin, i, :] += np.array(
                                [float(x) for x in t[1:]]
                            )
                            line = next(lines)
                            t = line.strip().split()
                            self.carray[ikpt, iband, iispin, i, :] += 1j * np.array(
                                [float(x) for x in t[1:]]
                            )
                        else:
                            raise Exception(
                                "Cannot parse line to projection: %s" % line
                            )



from .utilsprocar import UtilsProcar
from .procarparser import ProcarParser
from .procarselect import ProcarSelect
from .procarplot import ProcarPlot
from .procarsymmetry import ProcarSymmetry
from .fermisurface import FermiSurface
from .elkparser import ElkParser
from .abinitparser import AbinitParser
import matplotlib.pyplot as plt
from .splash import welcome


def fermi2D(
    file,
    outcar=None,
    abinit_output=None,
    spin=0,
    atoms=None,
    orbitals=None,
    energy=None,
    fermi=None,
    rec_basis=None,
    rot_symm=1,
    translate=[0, 0, 0],
    rotation=[0, 0, 0, 1],
    human=False,
    mask=None,
    savefig=None,
    st=False,
    noarrow=False,
    exportplt=False,
    code='vasp',
):
    """
  This module plots 2D Fermi surface.
  """

    welcome()

    # Turn interactive plotting off
    plt.ioff()

    if atoms is None:
        atoms = [-1]
        if human is True:
            print("WARNING: `--human` option given without atoms list!!!!!")

    if orbitals is None:
        orbitals = [-1]

    if rec_basis != None:
        rec_basis = np.array(rec_basis)
        rec_basis.shape = (3, 3)

    if len(translate) != 3 and len(translate) != 1:
        print("Error: --translate option is invalid! (", translate, ")")
        raise RuntimeError("invalid option --translate")

    print("file            : ", file)
    print("outcar          : ", outcar)
    print("Abinit output   : ", abinit_output)
    print("atoms           : ", atoms)
    print("orbitals        : ", orbitals)
    print("spin comp.      : ", spin)
    print("energy          : ", energy)
    print("fermi energy    : ", fermi)
    print("Rec. basis      : ", rec_basis)
    print("rot. symmetry   : ", rot_symm)
    print("origin (trasl.) : ", translate)
    print("rotation        : ", rotation)
    print("masking thres.  : ", mask)
    print("save figure     : ", savefig)
    print("st              : ", st)
    print("no_arrows       : ", noarrow)

    # first parse the outputs if given
    if code == 'vasp':
        if rec_basis is None and outcar:
            outcarparser = UtilsProcar()
            if fermi is None:
                fermi = outcarparser.FermiOutcar(outcar)
                print("Fermi energy found in outcar file = " + str(fermi))
            rec_basis = outcarparser.RecLatOutcar(outcar)
        # Reciprocal lattices are needed!
        elif rec_basis is None and outcar is None:
            print("ERROR: Reciprocal Lattice is needed, use --rec_basis or --outcar")
            raise RuntimeError("Reciprocal Lattice not found")

        # parsing the file
        procarFile = ProcarParser()
        # permissive incompatible with Fermi surfaces
        procarFile.readFile(file, permissive=False, recLattice=rec_basis)

    elif code == 'elk':
        procarFile = ElkParser()
        if rec_basis is None:
            if fermi is None:
                fermi = procarFile.fermi
                print("Fermi energy found in Elk output file = " + str(fermi))
            rec_basis = procarFile.reclat
        # Reciprocal lattices are needed!
        if rec_basis is None:
            print("ERROR: Reciprocal Lattice is needed, use --rec_basis or --outcar")
            raise RuntimeError("Reciprocal Lattice not found")
        procarFile = Elkparser(kdirect=False)

    elif code == 'abinit':
        if rec_basis is None and abinit_output:
            abinitparser = AbinitParser(abinit_output=abinit_output)
            if fermi is None:
                fermi = abinitparser.fermi
                print("Fermi energy found in Abinit ouput file = " + str(fermi))
            rec_basis = abinitparser.reclat
        # Reciprocal lattices are needed!
        elif rec_basis is None and abinit_output is None:
            print("ERROR: Reciprocal Lattice is needed, use --rec_basis or --outcar")
            raise RuntimeError("Reciprocal Lattice not found")
        # parsing the file
        procarFile = ProcarParser()
        # permissive incompatible with Fermi surfaces
        procarFile.readFile(file, permissive=False, recLattice=rec_basis)

    ### End of parsing ###

    if st is not True:
        # processing the data
        data = ProcarSelect(procarFile)
        data.selectIspin([spin])
        # fortran flag is equivalent to human,
        # but the later seems more human-friendly
        data.selectAtoms(atoms, fortran=human)
        data.selectOrbital(orbitals)
    else:
        # first get the sdp reduced array for all spin components.
        stData = []
        for i in [1, 2, 3]:
            data = ProcarSelect(procarFile)
            data.selectIspin([i])
            data.selectAtoms(atoms, fortran=human)
            data.selectOrbital(orbitals)
            stData.append(data.spd)

    # Once the PROCAR is parsed and reduced to 2x2 arrays, we can apply
    # symmetry operations to unfold the Brillouin Zone
    kpoints = data.kpoints
    bands = data.bands
    character = data.spd
    if st is True:
        sx, sy, sz = stData[0], stData[1], stData[2]
        symm = ProcarSymmetry(kpoints, bands, sx=sx, sy=sy, sz=sz, character=character)
    else:
        symm = ProcarSymmetry(kpoints, bands, character=character)

    symm.Translate(translate)
    symm.GeneralRotation(rotation[0], rotation[1:])
    # symm.MirrorX()
    symm.RotSymmetryZ(rot_symm)

    # plotting the data
    print("Bands will be shifted by the Fermi energy = ", fermi)
    fs = FermiSurface(symm.kpoints, symm.bands - fermi, symm.character)
    fs.FindEnergy(energy)

    if not st:
        fs.Plot(mask=mask, interpolation=300)
    else:
        fs.st(sx=symm.sx, sy=symm.sy, sz=symm.sz, noarrow=noarrow, spin=spin)

    if exportplt:
        return plt

    else:
        if savefig:
            plt.savefig(savefig, bbox_inches="tight")
            plt.close()  # Added by Nicholas Pike to close memory issue of looping and creating many figures
        else:
            plt.show()
        return



"""
Created on Fri March 31 2020
@author: Pedram Tavadze

"""
import numpy as np
import scipy.interpolate as interpolate
from .isosurface import Isosurface
from .brillouin_zone import BrillouinZone

class FermiSurface3D(Isosurface):
    def __init__(self,
                 kpoints=None,
                 band=None,
                 spd=None,
                 spd_spin=None,
                 fermi=None,
                 reciprocal_lattice=None,
                 interpolation_factor=1,
                 spin_texture=False,
                 color=None,
                 projection_accuracy='Normal',
                 cmap='viridis',
                 vmin=0,
                 vmax=1,
                 supercell=[1,1,1]
                 ):
        """
        

        Parameters
        ----------
        kpoints : TYPE, (n,3) float
            DESCRIPTION. A list of kpoints used in the DFT calculation, this list has to be
            (n,3), n being number of kpoints and 3 being the 3 different 
            cartesian coordinates
        band : TYPE, (n,) float
            DESCRIPTION. A list of energies of ith band cooresponding to the kpoints
        spd : TYPE, 
            DESCRIPTION. numpy array containing the information about ptojection of atoms,
            orbitals and spin on each band (check procarparser)
        fermi : TYPE, float
            DESCRIPTION. value of the fermi energy or any energy that one wants to find the 
            isosurface with 
        reciprocal_lattice : TYPE, (3,3)
            DESCRIPTION. The default is None. Reciprocal lattice of the structure
        interpolation_factor : TYPE, int
            DESCRIPTION. The default is 1. number of kpoints in every direction will increase by this factor
        color : TYPE, optional
            DESCRIPTION. The default is None.
        projection_accuracy : TYPE, optional
            DESCRIPTION. The default is 'Normal'.
        cmap : TYPE, string
            DESCRIPTION. The default is 'viridis'. Color map used in projecting the colors on the surface
        vmin : TYPE, float
            DESCRIPTION. The default is 0.
        vmax : TYPE, float
            DESCRIPTION. The default is 1.

        """






        self.kpoints = kpoints
        self.band = band
        self.spd = spd
        self.reciprocal_lattice = reciprocal_lattice
        self.supercell = supercell
        self.fermi = fermi
        self.interpolation_factor = interpolation_factor
        self.projection_accuracy = projection_accuracy
        self.spin_texture = spin_texture
        self.spd_spin = spd_spin
        
        self.brillouin_zone = self._get_brilloin_zone(supercell)

        Isosurface.__init__(self,
                            XYZ=self.kpoints,
                            V=self.band,
                            isovalue=self.fermi,
                            algorithm='lewiner',
                            interpolation_factor=interpolation_factor,
                            padding=self.supercell,
                            transform_matrix=self.reciprocal_lattice,
                            boundaries=self.brillouin_zone)
        if self.spd is not None and self.verts is not None:
            self.project_color(cmap, vmin, vmax)
        if self.spd_spin is not None and self.verts is not None:
            self.create_spin_texture()
            

    def create_spin_texture(self):
        
        if self.spd_spin is not None:
            XYZ_extended = self.XYZ.copy()
            vectors_extended_X = self.spd_spin[0].copy()
            vectors_extended_Y = self.spd_spin[1].copy()
            vectors_extended_Z = self.spd_spin[2].copy()

            for ix in range(3):
                for iy in range(self.supercell[ix]):
                    temp = self.XYZ.copy()
                    temp[:,ix]+=1*(iy+1)
                    XYZ_extended=np.append(XYZ_extended,temp,axis=0)
                    vectors_extended_X = np.append(vectors_extended_X,self.spd_spin[0],axis=0)
                    vectors_extended_Y = np.append(vectors_extended_Y,self.spd_spin[1],axis=0)
                    vectors_extended_Z = np.append(vectors_extended_Z,self.spd_spin[2],axis=0)
                    temp = self.XYZ.copy()
                    temp[:,ix]-=1*(iy+1)
                    XYZ_extended= np.append(XYZ_extended,temp,axis=0)
                    vectors_extended_X = np.append(vectors_extended_X,self.spd_spin[0],axis=0)
                    vectors_extended_Y = np.append(vectors_extended_Y,self.spd_spin[1],axis=0)
                    vectors_extended_Z = np.append(vectors_extended_Z,self.spd_spin[2],axis=0)


            # XYZ_extended = self.XYZ.copy()
            # scalars_extended = self.spd.copy()
            
            XYZ_transformed = np.dot(XYZ_extended, self.reciprocal_lattice)
            # XYZ_transformed = XYZ_extended
            
            if self.projection_accuracy.lower()[0] == 'n': #normal

                spin_X = interpolate.griddata(
                    XYZ_transformed, vectors_extended_X, self.verts, method="nearest"
                    )
                spin_Y = interpolate.griddata(
                    XYZ_transformed, vectors_extended_Y, self.verts , method="nearest"
                    )
                spin_Z = interpolate.griddata(
                    XYZ_transformed, vectors_extended_Z, self.verts , method="nearest"
                    )
                
            elif self.projection_accuracy.lower()[0] =='h': #high
                
                spin_X = interpolate.griddata(
                    XYZ_transformed, vectors_extended_X, self.verts, method="linear"
                    )
                spin_Y = interpolate.griddata(
                    XYZ_transformed, vectors_extended_Y, self.verts, method="linear"
                    )
                spin_Z = interpolate.griddata(
                    XYZ_transformed, vectors_extended_Z, self.verts, method="linear"
                    )

            self.set_vectors(spin_X,spin_Y,spin_Z)


            



    def project_color(self,cmap,vmin,vmax):
        """
        Projects the scalars to the surface.

        Parameters
        ----------
        cmap : TYPE string
            DESCRIPTION. Colormaps for the projection.
        vmin : TYPE
            DESCRIPTION.
        vmax : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        if self.spd is not None:
            XYZ_extended = self.XYZ.copy()
            scalars_extended = self.spd.copy()
            
            for ix in range(3):
                for iy in range(self.supercell[ix]):
                    temp = self.XYZ.copy()
                    temp[:,ix]+=1*(iy+1)
                    XYZ_extended=np.append(XYZ_extended,temp,axis=0)
                    scalars_extended = np.append(scalars_extended,self.spd,axis=0)
                    temp = self.XYZ.copy()
                    temp[:,ix]-=1*(iy+1)
                    XYZ_extended= np.append(XYZ_extended,temp,axis=0)
                    scalars_extended = np.append(scalars_extended,self.spd,axis=0)

            # XYZ_extended = self.XYZ.copy()
            # scalars_extended = self.spd.copy()
            
            XYZ_transformed = np.dot(XYZ_extended, self.reciprocal_lattice)
            # XYZ_transformed = XYZ_extended
            
            if self.projection_accuracy.lower()[0] == 'n': #normal
                colors = interpolate.griddata(
                    XYZ_transformed, scalars_extended, self.centers, method="nearest"
                    )
            elif self.projection_accuracy.lower()[0] =='h': #high
                colors = interpolate.griddata(
                        XYZ_transformed, scalars_extended, self.centers, method="linear"
                        )

            self.set_scalars(colors)
            self.set_color_with_cmap(cmap,vmin,vmax)



    def _get_brilloin_zone(self,supercell):
        return BrillouinZone(self.reciprocal_lattice,supercell)



# -*- coding: utf-8 -*-
"""
Created on Thu Jun 25 03:12:50 2020

@author: Logan Lang
"""


from re import findall, search, match, DOTALL, MULTILINE, finditer, compile, sub

from numpy import array, dot, linspace, sum, where, zeros, pi
import logging


class LobsterParser:
    def __init__(
        self,
        lobsterin="lobsterin",
        lobsterout="lobsterout",
        scfin="scf.in",
        outfile="scf.out",
        kdirect=True,
        lobstercode="qe",
    ):

        self.lobsterin = lobsterin
        self.scfin = scfin
        self.outfile = outfile
        self.lobsterout = lobsterout
        self.dosin = "DOSCAR.lobster"

        self.file_names = []

        self.discontinuities = []
        self.kdirect = kdirect
        ############################
        self.high_symmetry_points = None
        self.nhigh_sym = None
        self.nspecies = None

        self.composition = {}

        self.kticks = None
        self.knames = None

        self.speciesList = None

        self.kpoints = None
        self.kpointsCount = None
        self.bands = None
        self.bandsCount = None
        self.ionsList = None
        self.ionsCount = None
        self.spinCount = None
        
        # Used to store atomic states at the top of the kpdos.out file
        self.states = None

        self.test = None
        self.test2 = None

        self.spd = None
        """ for l=1:
              1 pz     (m=0)
              2 px     (real combination of m=+/-1 with cosine)
              3 py     (real combination of m=+/-1 with sine)
            for l=2:
              1 dz2    (m=0)
              2 dzx    (real combination of m=+/-1 with cosine)
              3 dzy    (real combination of m=+/-1 with sine)
              4 dx2-y2 (real combination of m=+/-2 with cosine)
              5 dxy    (real combination of m=+/-2 with sine)"""
        # Oribital order. This is the way it goes into the spd Array. index 0 is resered for totals
        self.orbitals = [
            "s",
            "p_y",
            "p_z",
            "p_x",
            "d_xy",
            "d_yz",
            "d_z^2",
            "d_x^2-y^2",
            "d_xz",
        ]

        # These are not used
        self.orbitalName_short = ["s", "p", "d", "tot"]
        self.orbitalCount = None

        # Opens the needed files
        rf = open(self.lobsterin, "r")
        self.lobsterin = rf.read()
        rf.close()

        rf = open(self.lobsterout, "r")
        self.lobsterout = rf.read()
        rf.close()

        rf = open(self.scfin, "r")
        self.scfin = rf.read()
        rf.close()
        self.dos = None

        self._readFileNames()
        self._readProjFiles()
        self._readHighKpoint()

        return

    ##########################################################################################
    # Finding high symmetry points
    ##########################################################################################

    def _readHighKpoint(self):
        numK = int(findall("K_POINTS.*\n([0-9]*)", self.scfin)[0])

        raw_khigh_sym = findall(
            "K_POINTS.*\n\s*[0-9]*.*\n" + numK * "(.*)\n", self.scfin,
        )[0]

        tickCountIndex = 0
        raw_high_symmetry = []
        self.knames = []
        self.kticks = []

        for x in raw_khigh_sym:
            if len(x.split()) == 5:

                raw_high_symmetry.append(
                    (float(x.split()[0]), float(x.split()[1]), float(x.split()[2]))
                )
                self.knames.append("$%s$" % x.split()[4].replace("!", ""))
                self.kticks.append(tickCountIndex)
            if float(x.split()[3]) == 0:
                tickCountIndex += 1
        self.high_symmetry_points = array(raw_high_symmetry)
        self.nhigh_sym = len(self.knames)

        # note:
        # discontinuities in Lobster are already specified in the input file.
        # The following is not needed.

        # Fixing discontunuities points if present.
        # Assume that they are separated by a '|'. E.g.- 'G|K'

        # dicontinuity_index = []
        # discontinuity_symbols = [s for s in self.knames if "|" in s]
        # for k in discontinuity_symbols:
        #     dicontinuity_index.append(self.knames.index(k))

        # for i in dicontinuity_index:
        #     self.discontinuities.append(self.kticks[i])

    ##########################################################################################
    # Finding file names
    ##########################################################################################

    def _readFileNames(self):

        self.ionsList = findall(
            "calculating FatBand for Element: (.*) Orbital.*", self.lobsterout
        )
        raw_speciesList = findall(
            "calculating FatBand for Element: ([a-zA-Z]*)[0-9] Orbital.*",
            self.lobsterout,
        )
        self.speciesList = []
        for x in raw_speciesList:
            if x not in self.speciesList:
                self.speciesList.append(x)
                self.composition.update({str(x): 0})
        self.nspecies = len(self.speciesList)
        for x in raw_speciesList:
            if x in self.speciesList:
                self.composition[x] += 1

        self.ionsCount = len(self.ionsList)

        proj_orbitals = findall(
            "calculating FatBand for Element: (.*) Orbital\(s\):\s*(.*)",
            self.lobsterout,
        )
        for proj in proj_orbitals:
            orbitals = proj[1].split()
            for orbital in orbitals:
                fileName = "FATBAND_" + proj[0] + "_" + orbital + ".lobster"
                self.file_names.append(fileName)

    ##########################################################################################
    # Reading projection files
    ##########################################################################################

    def _readProjFiles(self):
        rf = open(self.file_names[0], "r")
        projFile = rf.read()
        rf.close()

        ##########################################################################################
        # kpoints
        ##########################################################################################

        raw_kpoints = findall(
            "# K-Point \d+ :\s*([-\.\\d]*)\s*([-\.\\d]*)\s*([-\.\\d]*)", projFile
        )

        self.kpointsCount = len(raw_kpoints)
        self.kpoints = zeros(shape=(self.kpointsCount, 3))
        for ik in range(len(raw_kpoints)):
            for coord in range(3):
                self.kpoints[ik][coord] = raw_kpoints[ik][coord]
        # raw_kpoints = findall("(# K-Point \d+ :\s*[-\.\\d]*\s*[-\.\\d]*\s*[-\.\\d]*)", projFile)

        # If kdirect=False, then the kgrid will be in cartesian coordinates.
        # Requires the reciprocal lattice vectors to be parsed from the output.
        if not self.kdirect:
            self.kpoints = dot(self.kpoints, self.reclat)

        #########################################################################################
        # bands
        #########################################################################################

        raw_kpoints = findall(
            "# K-Point \d+ :\s*([-\.\\d]*\s*[-\.\\d]*\s*[-\.\\d]*)", projFile
        )
        self.bandsCount = int(findall("NBANDS\s*(\d*)", projFile)[0])

        band_info = []
        for ik in range(len(raw_kpoints)):
            expression = "# K-Point \d+ :\s*" + raw_kpoints[ik] + ".*\n"
            kpoint_bands = findall(expression + self.bandsCount * "(.*)\n", projFile)[0]
            for ikband in kpoint_bands:
                band_info.append(ikband)

#        if len(band_info) == self.bandsCount * self.kpointsCount:
#            print("Number of bands headers match")
        self.bands = zeros(shape=(self.kpointsCount, self.bandsCount))

        ik = 0
        ib = 0
        for i in range(len(band_info)):
            self.bands[ik, ib] = float(band_info[i].split()[1])
            ib += 1
            if int(ib == self.bandsCount):
                ik += 1
                ib = 0

        #########################################################################################
        # Forming SPD array
        #########################################################################################

        #Checks for spin polarization
        if len(findall("spillings for spin channel", self.lobsterout)) == 2:
            self.spinCount = 2
        else:
            self.spinCount = 1


        

        self.orbitalCount = 10
        self.spd = zeros(
            shape=(
                self.kpointsCount,
                self.bandsCount,
                self.spinCount,
                self.ionsCount + 1,
                len(self.orbitals) + 2,
            )
        )

        for file in range(len(self.file_names)):
            rf = open(self.file_names[file], "r")
            projFile = rf.read()
            rf.close()

            raw_kpoints = findall(
                "# K-Point \d+ :\s*([-\.\\d]*\s*[-\.\\d]*\s*[-\.\\d]*)", projFile
            )
            ionIndex = 0
            orbitalIndex = 0
            current_orbital = findall("(# FATBAND.*)", projFile)[0].split()[4]
            for i in range(len(self.ionsList)):
                if self.ionsList[i] == self.file_names[file].split("_")[1]:
                    ionIndex = i
                    # print(ionIndex)
            for i in range(len(self.orbitals)):
                if self.orbitals[i] == sub("[0-9]", "", current_orbital):
                    orbitalIndex = i + 1
                    # print(orbitalIndex)
                    # print(self.orbitals[orbitalIndex])

            band_info = []
            for ik in range(len(raw_kpoints)):
                expression = "# K-Point \d+ :\s*" + raw_kpoints[ik] + ".*\n"
                bands_wSpin = self.bandsCount * self.spinCount
                band = findall(expression + bands_wSpin * "(.*)\n", projFile)[0]
                for iband in range(self.bandsCount):
                    if self.spinCount == 2:
                        self.spd[ik,iband,0,ionIndex,orbitalIndex] += float(band[iband].split()[2])
                        self.spd[ik,iband,0,ionIndex,0] = ionIndex + 1 
                        self.spd[ik,iband,1,ionIndex,orbitalIndex] += float(band[iband+self.bandsCount].split()[2]) 
                        self.spd[ik,iband,1,ionIndex,0] = ionIndex + 1  
                    else:
                        self.spd[ik,iband,0,ionIndex,orbitalIndex] += float(band[iband].split()[2]) 
                        self.spd[ik,iband,0,ionIndex,0] = ionIndex + 1  
                    

            self.spd[:, :, :, :, -1] = sum(self.spd[:, :, :, :, 1:-1], axis=4)
            self.spd[:, :, :, -1, :] = sum(self.spd[:, :, :, 0:-1, :], axis=3)
            self.spd[:, :, :, -1, 0] = 0
            
        # colinear spin polarized case
        
        # manipulating spd array for spin polarized calculations.
        # The shape is (nkpoints,2*nbands,2,natoms,norbitals)
        # The third dimension is for spin.
        # When this is zero, the bands*2 (all spin up and down bands) have positive projections.
        # When this is one, the the first half of bands (spin up) will have positive projections
        # and the second half (spin down) will have negative projections. This is to adhere to
        # the convention used in PyProcar to obtain spin density and spin magnetization.
    
        if self.spinCount == 2:
            print("\nLobster colinear spin calculation detected.\n")
            self.spd2 = zeros(
            shape=(
                self.kpointsCount,
                self.bandsCount*2,
                self.spinCount,
                self.ionsCount + 1,
                len(self.orbitals) + 2,
                )
            )
            
            # spin up block for spin=0
            self.spd2[:, :self.bandsCount, 0, :, :] = self.spd[:,:,0,:,:]
            
            # spin down block for spin=0
            self.spd2[:, self.bandsCount:, 0, :, :] = self.spd[:,:,1,:,:]
            
            # spin up block for spin=1
            self.spd2[:, :self.bandsCount, 1, :, :] = self.spd[:,:,0,:,:]
            
            # spin down block for spin=1
            self.spd2[:, self.bandsCount:, 1, :, :] = -1*self.spd[:,:,1,:,:]
            
            self.spd = self.spd2

    @property
    def fermi(self):
        """
        Returns the fermi energy read from .out
        """

        fi = open(self.outfile, "r")
        data = fi.read()
        fi.close()
        fermi = float(findall(r"the\s*Fermi\s*energy\s*is\s*([\s\d.]*)ev", data)[0])
        return fermi

    @property
    def reclat(self):
        """
        Returns the reciprocal lattice read from .out
        """
        rf = open(self.outfile, "r")
        data = rf.read()
        rf.close()

        alat = float(findall(r"alat\)\s*=\s*([\d.e+-]*)", data)[0])

        b1 = array(
            findall(r"b\(1\)\s*=\s*\(([\d.\s+-e]*)", data)[0].split(), dtype="float64"
        )
        b2 = array(
            findall(r"b\(2\)\s*=\s*\(([\d.\s+-e]*)", data)[0].split(), dtype="float64"
        )
        b3 = array(
            findall(r"b\(3\)\s*=\s*\(([\d.\s+-e]*)", data)[0].split(), dtype="float64"
        )

        reclat = (2 * pi / alat) * (array((b1, b2, b3)))

        # Transposing to get the correct format
        reclat = reclat.T

        return reclat


"""
Created on May 17 2020
@author: Pedram Tavadze
"""
from .elkparser import ElkParser
from .splash import welcome
from .doscarplot import DosPlot
import numpy as np
import matplotlib.pyplot as plt


# import matplotlib
plt.rcParams["mathtext.default"] = "regular"  # Roman ['rm', 'cal', 'it', 'tt', 'sf',
#                                                   'bf', 'default', 'bb', 'frak',
#                                                   'circled', 'scr', 'regular']
plt.rcParams["font.family"] = "Georgia"
plt.rc("font", size=18)  # controls default text sizes
plt.rc("axes", titlesize=22)  # fontsize of the axes title
plt.rc("axes", labelsize=22)  # fontsize of the x and y labels
plt.rc("xtick", labelsize=22)  # fontsize of the tick labels
plt.rc("ytick", labelsize=22)  # fontsize of the tick labels
# plt.rc('legend', fontsize=22)    # legend fontsize
# plt.rc('figure', titlesize=22)  # fontsize of the figure title


def dosplot(
    vasprunfile="vasprun.xml",
    mode="plain",
    orientation="horizontal",
    spin_colors=None,
    colors=None,
    spins=None,
    atoms=None,
    orbitals=None,
    elimit=None,
    cmap="jet",
    vmax=None,
    vmin=None,
    grid=True,
    savefig=None,
    title=None,
    plot_total=True,
    code="vasp",
    labels=None,
    items={},
    ax=None,
):

    """This function plots density of states

    """

    welcome()

    # Verbose section
    print("Script initiated")
    print("code          : ", code)
    print("vasprun file  : ", vasprunfile)
    print("mode          : ", mode)
    print("spin          : ", spins)
    print("atoms list    : ", atoms)
    print("orbs. list    : ", orbitals)
    print("energy range  : ", elimit)
    print("eolormap      : ", cmap)
    print("vmax          : ", vmax)
    print("vmin          : ", vmin)
    print("grid enabled  : ", grid)
    print("savefig       : ", savefig)
    print("title         : ", title)

    total = plot_total
    code = code.lower()
    if code == "vasp":
        dos_plot = DosPlot(vasprunfile)
        vaspxml = dos_plot.VaspXML
        if atoms is None:
            atoms = list(np.arange(vaspxml.initial_structure.natom, dtype=int))
        if spins is None:
            spins = list(np.arange(vaspxml.dos_total.ncols))
        if orbitals is None:
            orbitals = list(
                np.arange((len(vaspxml.dos_projected[0].labels) - 1) // 2, dtype=int)
            )
        if elimit is None:
            elimit = [
                vaspxml.dos_total.energies.min(),
                vaspxml.dos_total.energies.max(),
            ]

    if mode == "plain":
        fig, ax1 = dos_plot.plot_total(
            spins=spins,
            spin_colors=spin_colors,
            ax=ax,
            orientation=orientation,
            labels=labels,
        )
        dos = dos_plot.VaspXML.dos_total

    elif mode == "parametric_line":
        if not total:
            fig, ax1 = dos_plot.plot_parametric_line(
                atoms=atoms,
                spins=spins,
                orbitals=orbitals,
                spin_colors=spin_colors,
                ax=ax,
                orientation=orientation,
                labels=labels,
            )
            dos = dos_plot.VaspXML.dos_parametric(
                atoms=atoms, spin=spins, orbitals=orbitals,
            )
        else:
            fig, ax1 = dos_plot.plot_total(
                spins=spins,
                spin_colors=[(0, 0, 0), (0, 0, 0)],
                ax=ax,
                orientation=orientation,
            )
            dos = dos_plot.VaspXML.dos_total
            _, ax1 = dos_plot.plot_parametric_line(
                atoms=atoms,
                spins=spins,
                orbitals=orbitals,
                spin_colors=spin_colors,
                ax=ax1,
                orientation=orientation,
                labels=labels,
            )
    elif mode == "parametric":
        if not total:
            fig, ax1 = dos_plot.plot_parametric(
                atoms=atoms,
                spins=spins,
                orbitals=orbitals,
                spin_colors=spin_colors,
                cmap=cmap,
                elimit=elimit,
                ax=ax,
                orientation=orientation,
                labels=labels,
            )
            dos = dos_plot.VaspXML.dos_total

        else:
            fig, ax1 = dos_plot.plot_parametric(
                atoms=atoms,
                spins=spins,
                orbitals=orbitals,
                spin_colors=spin_colors,
                cmap=cmap,
                elimit=elimit,
                ax=ax,
                orientation=orientation,
                labels=labels,
            )
            dos = dos_plot.VaspXML.dos_total
            _, ax1 = dos_plot.plot_total(
                spins=spins,
                spin_colors=[(0, 0, 0), (0, 0, 0)],
                ax=ax1,
                orientation=orientation,
            )

    elif mode == "stack_species":
        if not total:
            fig, ax1 = dos_plot.plot_stack_species(
                spins=spins,
                orbitals=orbitals,
                spin_colors=spin_colors,
                colors=colors,
                elimit=elimit,
                figsize=(12, 6),
                ax=ax,
                orientation=orientation,
            )
            dos = dos_plot.VaspXML.dos_total
        else:
            fig, ax1 = dos_plot.plot_stack_species(
                spins=spins,
                orbitals=orbitals,
                spin_colors=spin_colors,
                colors=colors,
                elimit=elimit,
                figsize=(12, 6),
                ax=ax,
                orientation=orientation,
            )
            dos = dos_plot.VaspXML.dos_total
            _, ax1 = dos_plot.plot_total(
                spins=spins,
                spin_colors=[(0, 0, 0), (0, 0, 0)],
                ax=ax1,
                orientation=orientation,
            )

    elif mode == "stack_orbitals":
        if not total:
            fig, ax1 = dos_plot.plot_stack_orbitals(
                spins=spins,
                atoms=atoms,
                spin_colors=spin_colors,
                colors=colors,
                elimit=elimit,
                figsize=(12, 6),
                ax=ax,
                orientation=orientation,
            )
            dos = dos_plot.VaspXML.dos_total
        else:
            fig, ax1 = dos_plot.plot_stack_orbitals(
                spins=spins,
                atoms=atoms,
                spin_colors=spin_colors,
                colors=colors,
                elimit=elimit,
                figsize=(12, 6),
                ax=ax,
                orientation=orientation,
            )
            dos = dos_plot.VaspXML.dos_total
            _, ax1 = dos_plot.plot_total(
                spins=spins,
                spin_colors=[(0, 0, 0), (0, 0, 0)],
                ax=ax1,
                orientation=orientation,
            )

    elif mode == "stack":
        if not total:
            fig, ax1 = dos_plot.plot_stack(
                items=items,
                spins=spins,
                spin_colors=spin_colors,
                colors=colors,
                elimit=elimit,
                figsize=(12, 6),
                ax=ax,
                orientation=orientation,
            )

            dos = dos_plot.VaspXML.dos_total
        else:
            fig, ax1 = dos_plot.plot_stack(
                items=items,
                spins=spins,
                spin_colors=colors,
                colors=colors,
                elimit=elimit,
                figsize=(12, 6),
                ax=ax,
                orientation=orientation,
            )
            dos = dos_plot.VaspXML.dos_total
            _, ax1 = dos_plot.plot_total(
                spins=spins,
                spin_colors=[(0, 0, 0), (0, 0, 0)],
                ax=ax1,
                orientation=orientation,
            )

    cond1 = dos.energies > elimit[0]
    cond2 = dos.energies < elimit[1]
    cond = np.all([cond1, cond2], axis=0)

    if len(spins) > 1:
        ylim = [dos.values[cond][:, 1].max() * -1.1, dos.values[cond][:, 0].max() * 1.1]
    else:
        ylim = [0, dos.dos[cond][:, spins[0] + 1].max() * 1.1]

    if orientation == "horizontal":
        ax1.set_xlabel(r"$E-E_f$ [eV]")
        ax1.set_ylabel("Density of States [a.u.]")
        ax1.set_xlim(elimit)
        ax1.set_ylim(ylim)

    elif orientation == "vertical":
        ax1.set_ylabel(r"$E-E_f$ [eV]")
        ax1.set_xlabel("Density of States [a.u.]")
        ax1.set_ylim(elimit)
        ax1.set_xlim(ylim)  # we use ylim because the plot is vertiacal

    ax1.axhline(color="black", linestyle="--")
    ax1.axvline(color="black", linestyle="--")

    fig.tight_layout()
    if grid:
        ax1.grid()
    if labels or "stack" in mode:
        ax1.legend()
    if title:
        ax1.set_title(title, fontsize=17)

    if savefig:

        fig.savefig(savefig, bbox_inches="tight")
        plt.close()  # Added by Nicholas Pike to close memory issue of looping and creating many figures
        return None, None
    else:
        plt.show()

    return fig, ax1


#
#
## if __name__ == "__main__":
## bandsplot(mode='parametric',elimit=[-6,6],orbitals=[4,5,6,7,8],vmin=0,vmax=1, code='elk')
## knames=['$\Gamma$', '$X$', '$M$', '$\Gamma$', '$R$','$X$'],
## kticks=[0, 8, 16, 24, 38, 49])



from .utilsprocar import UtilsProcar
from .splash import welcome


# calls ProcarRepair
def repair(infile, outfile):
    """
	This module calls ProcarRepair to repair the PROCAR file.
	"""
    welcome()

    print("Input File    : ", infile)
    print("Output File   : ", outfile)

    # parsing the file
    handler = UtilsProcar()
    handler.ProcarRepair(infile, outfile)



__author__ = "Francisco Munoz,Aldo Romero,Sobhit Singh,Uthpala Herath,Pedram Tavadze,Eric Bousquet,Xu He"
__copyright__ = "Copyright 2020"
__version__ = "5.5.1"
__email__ = (
    "fvmunoz@gmail.com/alromero@mail.wvu.edu/ukh0001@mix.wvu.edu/petavazohi@mix.wvu.edu"
)
__status__ = "Development"
__date__ = "July 27th, 2020"

# Copyright (C) 2020 Francisco Munoz,Aldo Romero,Sobhit Singh,Uthpala Herath,Pedram Tavadze,Eric Bousquet,Xu He
#
# This file is part of pyprocar.
#
# PyProcar is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyProcar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyprocar.  If not, see <http://www.gnu.org/licenses/>.


import logging
import re
import sys

import ase
import matplotlib.pyplot as plt
import numpy as np
import seekpath
import skimage
import pychemia
import pyvista
import trimesh

# Abinit
from pyprocar.abinitparser import AbinitParser

# Elk
from pyprocar.elkparser import ElkParser

# Quantum Espresso
from pyprocar.qeparser import QEParser

# Lobster
from pyprocar.lobsterparser import LobsterParser

from pyprocar.fermisurface import FermiSurface
from pyprocar.fermisurface3d import FermiSurface3D
from . import fermisurface3d
from pyprocar.procarfilefilter import ProcarFileFilter
from pyprocar.procarparser import ProcarParser
from pyprocar.procarplot import ProcarPlot
from pyprocar.procarselect import ProcarSelect
from pyprocar.procarsymmetry import ProcarSymmetry
from pyprocar.utilsprocar import UtilsProcar
from pyprocar.doscarplot import DosPlot

from .scriptBandsDosplot import bandsdosplot
from .scriptBandsplot import bandsplot
from .scriptCat import cat
from .scriptDosplot import dosplot
from .scriptFermi2D import fermi2D
from .scriptFermi3D import fermi3D
from .scriptFilter import filter
from .scriptKmesh2D import generate2dkmesh
from .scriptKpath import kpath
from .scriptRepair import repair
from .scriptUnfold import unfold
from .scriptVector import Vector
from .scriptBandGap import bandgap
from .splash import welcome



import pyprocar


def welcome():
    print(
        " ____        ____\n|  _ \ _   _|  _ \ _ __ ___   ___ __ _ _ __ \n| |_) | | | | |_) | '__/ _ \ / __/ _` | '__|\n|  __/| |_| |  __/| | | (_) | (_| (_| | |   \n|_|    \__, |_|   |_|  \___/ \___\__,_|_|\n       |___/"
    )
    print("A Python library for electronic structure pre/post-processing.\n")
    print("Version %s created on %s\n" % (pyprocar.__version__, pyprocar.__date__))
    print(
        "Please cite: Herath, U., Tavadze, P., He, X., Bousquet, E., Singh, S., Muñoz, F. & Romero,\
    A., PyProcar: A Python library for electronic structure pre/post-processing.,\
    Computer Physics Communications 251 (2020):107080.\n"
    )

    return



import numpy as np
import re
import logging
import matplotlib.pyplot as plt
import sys


class ProcarPlot:
    def __init__(self, bands, spd, kpoints=None):
        self.bands = bands.transpose()
        self.spd = spd.transpose()
        self.kpoints = kpoints
        return

    def plotBands(
        self,
        size=0.02,
        marker="o",
        ticks=None,
        color="blue",
        discontinuities=[],
        figsize=(13, 9),
        ax=None,
    ):

        if not ax:
            fig = plt.figure(figsize=figsize)
            fig.tight_layout()
            ax = fig.add_subplot(111)
        else:
            fig = plt.gca()

        if size is not None:
            size = size / 2

        if self.kpoints is not None:
            xaxis = [0]

            #### MODIFIED FOR DISCONTINOUS BANDS ####
            if ticks:
                ticks, ticksNames = list(zip(*ticks))

                # counters for number of discontinuities
                icounter = 1
                ii = 0

                for i in range(1, len(self.kpoints) - len(discontinuities)):
                    d = self.kpoints[icounter] - self.kpoints[icounter - 1]
                    d = np.sqrt(np.dot(d, d))
                    xaxis.append(d + xaxis[-1])
                    icounter += 1
                    ii += 1
                    if ii in discontinuities:
                        icounter += 1
                        ii += 1
                        xaxis.append(xaxis[-1])
                xaxis = np.array(xaxis)

                # plotting

                for i_tick in range(len(ticks) - 1):
                    x = xaxis[ticks[i_tick] : ticks[i_tick + 1] + 1]
                    y = self.bands.transpose()[ticks[i_tick] : ticks[i_tick + 1] + 1, :]
                    ax.plot(x, y, "r-", marker=marker, markersize=size, color=color)

            #### END  OF MODIFIED DISCONTINUOUS BANDS ####

            # if ticks are not given
            else:
                for i in range(1, len(self.kpoints)):
                    d = self.kpoints[i - 1] - self.kpoints[i]
                    d = np.sqrt(np.dot(d, d))
                    xaxis.append(d + xaxis[-1])
                xaxis = np.array(xaxis)
                ax.plot(
                    xaxis,
                    self.bands.transpose(),
                    "r-",
                    marker=marker,
                    markersize=size,
                    color=color,
                )

        # if self.kpoints is None
        else:
            xaxis = np.arange(len(self.bands))
            ax.plot(
                xaxis,
                self.bands.transpose(),
                "r-",
                marker=marker,
                markersize=size,
                color=color,
            )

        ax.set_xlim(xaxis.min(), xaxis.max())

        # Handling ticks
        if ticks:
            # added for meta-GGA calculations
            if ticks[0] > 0:
                ax.set_xlim(left=xaxis[ticks[0]])
            ticks = [xaxis[x] for x in ticks]
            ax.set_xticks(ticks)
            ax.set_xticklabels(ticksNames)
            for xc in ticks:
                ax.axvline(x=xc, color="k")
        ax.axhline(color="black", linestyle="--")

        return fig, ax

    def parametricPlot(
        self,
        cmap="jet",
        vmin=None,
        vmax=None,
        mask=None,
        ticks=None,
        discontinuities=[],
        ax=None,
        figsize=(13, 9),
        plot_bar=True,
    ):
        from matplotlib.collections import LineCollection
        import matplotlib

        # fig = plt.figure() # use plt.gca() since it won't create a new figure for band comparison
        if ax is None:
            fig = plt.figure(figsize=figsize)
            fig.tight_layout()
            ax = fig.add_subplot(111)
        else:
            fig = ax.get_figure()
        bsize, ksize = self.bands.shape

        # print self.bands
        if mask is not None:
            mbands = np.ma.masked_array(self.bands, np.abs(self.spd) < mask)
        else:
            # Faking a mask, all elemtnet are included
            mbands = np.ma.masked_array(self.bands, False)

        if vmin is None:
            vmin = self.spd.min()
        if vmax is None:
            vmax = self.spd.max()
        print("normalizing to : ", (vmin, vmax))
        norm = matplotlib.colors.Normalize(vmin, vmax)

        if self.kpoints is not None:
            xaxis = [0]

            #### MODIFIED FOR DISCONTINOUS BANDS ####
            if ticks:
                ticks, ticksNames = list(zip(*ticks))

                # counters for number of discontinuities
                icounter = 1
                ii = 0

                for i in range(1, len(self.kpoints) - len(discontinuities)):
                    d = self.kpoints[icounter] - self.kpoints[icounter - 1]
                    d = np.sqrt(np.dot(d, d))
                    xaxis.append(d + xaxis[-1])
                    icounter += 1
                    ii += 1
                    if ii in discontinuities:
                        icounter += 1
                        ii += 1
                        xaxis.append(xaxis[-1])
                xaxis = np.array(xaxis)

                # plotting
                for y, z in zip(mbands, self.spd):
                    points = np.array([xaxis, y]).T.reshape(-1, 1, 2)
                    segments = np.concatenate([points[:-1], points[1:]], axis=1)
                    lc = LineCollection(segments, cmap=plt.get_cmap(cmap), norm=norm)
                    lc.set_array(z)
                    lc.set_linewidth(1)
                    ax.add_collection(lc)
                if plot_bar:
                    cb = fig.colorbar(lc, ax=ax)
                    cb.ax.tick_params(labelsize=20)
                ax.set_xlim(xaxis.min(), xaxis.max())
                ax.set_ylim(mbands.min(), mbands.max())

            #### END  OF MODIFIED DISCONTINUOUS BANDS ####

            # if ticks are not given
            else:
                for i in range(1, len(self.kpoints)):
                    d = self.kpoints[i - 1] - self.kpoints[i]
                    d = np.sqrt(np.dot(d, d))
                    xaxis.append(d + xaxis[-1])
                xaxis = np.array(xaxis)

                # plotting
                for y, z in zip(mbands, self.spd):
                    points = np.array([xaxis, y]).T.reshape(-1, 1, 2)
                    segments = np.concatenate([points[:-1], points[1:]], axis=1)
                    lc = LineCollection(segments, cmap=plt.get_cmap(cmap), norm=norm)
                    lc.set_array(z)
                    lc.set_linewidth(1)
                    ax.add_collection(lc)
                if plot_bar:
                    cb = fig.colorbar(lc, ax=ax)
                    cb.ax.tick_params(labelsize=20)
                ax.set_xlim(xaxis.min(), xaxis.max())
                ax.set_ylim(mbands.min(), mbands.max())

        # if self.kpoints is None
        else:
            xaxis = np.arange(ksize)
            for y, z in zip(mbands, self.spd):
                points = np.array([xaxis, y]).T.reshape(-1, 1, 2)
                segments = np.concatenate([points[:-1], points[1:]], axis=1)
                lc = LineCollection(segments, cmap=plt.get_cmap(cmap), norm=norm)
                lc.set_array(z)
                lc.set_linewidth(1)
                ax.add_collection(lc)
            if plot_bar:
                cb = fig.colorbar(lc, ax=ax)
                cb.ax.tick_params(labelsize=20)
            ax.set_xlim(xaxis.min(), xaxis.max())
            ax.set_ylim(mbands.min(), mbands.max())

        # handling ticks
        if ticks:
            # added for meta-GGA calculations
            if ticks[0] > 0:
                ax.set_xlim(left=xaxis[ticks[0]])
            ticks = [xaxis[x] for x in ticks]
            ax.set_xticks(ticks)
            ax.set_xticklabels(ticksNames)
            for xc in ticks:
                ax.axvline(x=xc, color="black")

        ax.axhline(color="black", linestyle="--")

        return fig, ax

    def scatterPlot(
        self,
        size=50,
        mask=None,
        cmap="hot_r",
        vmax=None,
        vmin=None,
        marker="o",
        ticks=None,
        discontinuities=[],
        ax=None,
    ):
        bsize, ksize = self.bands.shape
        # plotting
        if not ax:
            fig = plt.figure(figsize=(13, 9))
            fig.tight_layout()
            ax = fig.add_subplot(111)
        else:
            fig = ax.get_figure()
        if self.kpoints is not None:
            xaxis = [0]

            #### MODIFIED FOR DISCONTINOUS BANDS ####
            if ticks:
                ticks, ticksNames = list(zip(*ticks))

                # counters for number of discontinuities
                icounter = 1
                ii = 0

                for i in range(1, len(self.kpoints) - len(discontinuities)):
                    d = self.kpoints[icounter] - self.kpoints[icounter - 1]
                    d = np.sqrt(np.dot(d, d))
                    xaxis.append(d + xaxis[-1])
                    icounter += 1
                    ii += 1
                    if ii in discontinuities:
                        icounter += 1
                        ii += 1
                        xaxis.append(xaxis[-1])
                xaxis = np.array(xaxis)

                xaxis.shape = (1, ksize)
                xaxis = xaxis.repeat(bsize, axis=0)
                if mask is not None:
                    mbands = np.ma.masked_array(self.bands, np.abs(self.spd) < mask)
                else:
                    mbands = self.bands

                scatter = ax.scatter(
                    xaxis,
                    mbands,
                    c=self.spd,
                    s=size,
                    linewidths=0,
                    cmap=cmap,
                    vmax=vmax,
                    vmin=vmin,
                    marker=marker,
                    edgecolors="none",
                )
                fig.colorbar(scatter)
                ax.set_xlim(xaxis.min(), xaxis.max())

            #### END  OF MODIFIED DISCONTINUOUS BANDS ####

            # if ticks are not given
            else:
                for i in range(1, len(self.kpoints)):
                    d = self.kpoints[i - 1] - self.kpoints[i]
                    d = np.sqrt(np.dot(d, d))
                    xaxis.append(d + xaxis[-1])
                xaxis = np.array(xaxis)

                xaxis.shape = (1, ksize)
                xaxis = xaxis.repeat(bsize, axis=0)
                if mask is not None:
                    mbands = np.ma.masked_array(self.bands, np.abs(self.spd) < mask)
                else:
                    mbands = self.bands

                scatter = ax.scatter(
                    xaxis,
                    mbands,
                    c=self.spd,
                    s=size,
                    linewidths=0,
                    cmap=cmap,
                    vmax=vmax,
                    vmin=vmin,
                    marker=marker,
                    edgecolors="none",
                )

                plt.colorbar(scatter)
                ax.set_xlim(xaxis.min(), xaxis.max())

        # if kpoints is None
        else:
            xaxis = np.arange(ksize)
            xaxis.shape = (1, ksize)
            xaxis = xaxis.repeat(bsize, axis=0)
            if mask is not None:
                mbands = np.ma.masked_array(self.bands, np.abs(self.spd) < mask)
            else:
                mbands = self.bands

            ax.scatter(
                xaxis,
                mbands,
                c=self.spd,
                s=size,
                linewidths=0,
                cmap=cmap,
                vmax=vmax,
                vmin=vmin,
                marker=marker,
                edgecolors="none",
            )

            plt.colorbar(ax=ax)
            ax.set_xlim(xaxis.min(), xaxis.max())

        # handling ticks
        if ticks:
            # added for meta-GGA calculations
            if ticks[0] > 0:
                ax.set_xlim(left=xaxis[0, ticks[0]])
            ticks = [xaxis[0, x] for x in ticks]
            ax.set_xticks(ticks)
            ax.set_xticklabels(ticksNames)

        ax.axhline(color="black", linestyle="--")

        return fig, ax

    def atomicPlot(self, cmap="hot_r", vmin=None, vmax=None, ax=None):
        """
    Just a handler to parametricPlot. Useful to plot energy levels.

    It adds a fake k-point. Shouldn't be invoked with more than one
    k-point
    ax not implemented here, not need
    """

        print("Atomic plot: bands.shape  :", self.bands.shape)
        print("Atomic plot: spd.shape    :", self.spd.shape)
        print("Atomic plot: kpoints.shape:", self.kpoints.shape)

        self.bands = np.hstack((self.bands, self.bands))
        self.spd = np.hstack((self.spd, self.spd))
        self.kpoints = np.vstack((self.kpoints, self.kpoints))
        self.kpoints[0][-1] += 1
        print("Atomic plot: bands.shape  :", self.bands.shape)
        print("Atomic plot: spd.shape    :", self.spd.shape)
        print("Atomic plot: kpoints.shape:", self.kpoints.shape)

        #        print(self.kpoints)

        fig, ax1 = self.parametricPlot(cmap, vmin, vmax, ax=ax)

        #        plt.gca().xaxis.set_major_locator(plt.NullLocator())
        ax1.xaxis.set_major_locator(plt.NullLocator())
        # labels on each band
        for i in range(len(self.bands[:, 0])):
            # print i, self.bands[i]
            ax1.text(0, self.bands[i, 0], str(i + 1))

        return fig, ax1



from .procarfilefilter import ProcarFileFilter
from .splash import welcome


def filter(
    inFile,
    outFile,
    atoms=None,
    orbitals=None,
    orbital_names=None,
    bands=None,
    spin=None,
    kpoints=None,
    human_atoms=False,
):
    """
    This module filters the PROCAR file and re-write a new one.
    """
    welcome()

    print("Input file  :", inFile)
    print("Output file :", outFile)

    print("atoms       :", atoms)
    if atoms:
        print("human_atoms     :", human_atoms)
    print("orbitals  :", orbitals)
    if orbitals:
        print("orb. names  :", orbital_names)
    print("bands       :", bands)
    print("spins       :", spin)
    print("k-points    :", kpoints)

    # Access init class of ProcarFileFilter and pass two arguments
    FileFilter = ProcarFileFilter(inFile, outFile)

    # for atoms
    if atoms:
        print("Manipulating the atoms")

        if human_atoms:
            atoms = [[y - 1 for y in x] for x in atoms]
            print("new atoms list :", atoms)

        # Now just left to call the driver member
        FileFilter.FilterAtoms(atoms)

    # for orbitals
    elif orbitals:
        print("Manipulating the orbitals")
        # If orbitals orbital_names is None, it needs to be filled
        if orbital_names is None:
            orbital_names = ["o" + str(x) for x in range(len(orbitals))]
            print("New orbitals names (default): ", orbital_names)
        # testing if makes sense
        if len(orbitals) != len(orbital_names):
            raise RuntimeError("length of orbitals and orbitals names do not match")

        FileFilter.FilterOrbitals(orbitals, orbital_names)

    # for bands
    elif bands:
        print("Manipulating the bands")

        bmin = bands[0]
        bmax = bands[1]
        if bmax < bmin:
            bmax, bmin = bmin, bmax
            print("New bands limits: ", bmin, " to ", bmax)

        FileFilter.FilterBands(bmin, bmax)

    # for k-points
    elif kpoints:
        print("Manipulating the k-points")

        kmin = kpoints[0]
        kmax = kpoints[1]
        if kmax < kmin:
            kmax, kmin = kmin, kmax
            print("New k-points limits: ", kmin, " to ", kmax)

        FileFilter.FilterKpoints(kmin, kmax)

    # for spin
    elif spin:
        print("Manipulating the spin")

        FileFilter.FilterSpin(spin)

    return



import logging
import re
import sys

import matplotlib.pyplot as plt
import numpy as np

from ..utilsprocar import UtilsProcar


class ProcarSelect:
    """
  Reduces the dimensionality of the data making it uselful to
  plot bands.

  The main data to manipulate is the projected electronic structure.
  Its shape original is:

  spd[kpoint][band][ispin][atom][orbital].

  The selection of components should be done in order, says, first
  "ispin", then "atom", and at last "orbital".

  Note: once any selection has been performed, the data itself
  changes. Say, if you want compare atom [0] and [1,2], you need two
  instances of this class.


  Example to compare the bandstructure of two set of atoms
  >>>

  """

    def __init__(
        self, ProcarData=None, deepCopy=True, loglevel=logging.WARNING, mode=None
    ):

        self.spd = None
        self.bands = None
        self.kpoints = None
        # self.cspd=None
        self.mode = mode

        # We want a logging to tell us what is happening
        self.log = logging.getLogger("ProcarSelect")
        self.log.setLevel(loglevel)
        self.ch = logging.StreamHandler()
        self.ch.setFormatter(
            logging.Formatter("%(name)s::%(levelname)s:" " %(message)s")
        )
        self.ch.setLevel(logging.DEBUG)
        self.log.addHandler(self.ch)
        # At last, one message to the logger.
        self.log.debug("ProcarSelect: instanciated")

        if ProcarData is not None:
            self.setData(ProcarData, deepCopy)
        return

    def setData(self, ProcarData, deepCopy=True):
        """
    The data from ProcarData is deepCopy-ed by default (ie: their
    elements are not modified by this class.

    Args:

    -ProcarData: is a ProcarParser instance (or anything with similar
     functionality, duck typing)

    -deepCopy=True: If false a shallow copy will be made (saves memory).
    """
        self.log.debug("setData: ...")
        if deepCopy is True:
            self.spd = ProcarData.spd.copy()
            self.bands = ProcarData.bands.copy()
            self.kpoints = ProcarData.kpoints.copy()
            # self.cspd = ProcarData.cspd.copy()
        else:
            self.spd = ProcarData.spd
            self.bands = ProcarData.bands
            self.kpoints = ProcarData.kpoints
            # self.cspd = ProcarData.cspd

        self.numspin = self.spd.shape[2]  # Number of spins
        self.log.debug("setData: ... Done")
        print(
            "spd shape      :  %s [kpoints, bands, spins, atoms+1, orbitals+2]"
            % str(self.spd.shape)
        )

        return

    def selectIspin(self, value=None, separate=False):
        """
    value is a list with the values of Ispin to select.

    UPDATE:
        if separate == true, then spin = 0 corresponds to spin up
        and spin = 1 corresponds to spin down. If not, they give
        spin density and spin magnetization, respectively.

    Example:
    >>> foo = ProcarParser()
    >>> foo.readFile("PROCAR")
    >>> bar = ProcarSelect(foo)
    >>> bar.selectIspin([0]) #just the density
    """
        # all kpoint, all bands, VALUE spin, all the rest
        self.log.debug("selectIspin: ...")
        self.log.debug("old spd shape =" + str(self.spd.shape))
        # first, testing if the domensionaluty is rigth:
        dimen = len(self.spd.shape)
        if dimen != 5:
            self.log.error(
                "The array is " + str(dimen) + " dimensional, expecting a"
                " 5 dimensional array."
            )
            self.log.error(
                "You should call selectIspin->selecAtom->selectOrbitals, "
                "in this order."
            )
            raise RuntimeError("Wrong dimensionality of the array")
        self.log.debug("ispin value = " + str(value))

        numofbands = int(self.spd.shape[1] / 2)

        if separate == False:
            # spin density or magnetization
            self.spd = self.spd[:, :, value]
            self.spd = self.spd.sum(axis=2)
            self.log.info("new spd shape =" + str(self.spd.shape))
            self.log.debug("selectIspin: ...Done")
            if self.mode == "parametric":
                if self.numspin > 1:
                    if value == [0]:
                        print("Plotting spin density...")
                    elif value == [1]:
                        print("Plotting spin magnetization...")

        else:
            # spin up (spin = 0) and spin down (spin = 1) separately.
            if value == [0]:
                # select spin up block
                self.spd = self.spd[:, :numofbands, 0]

            elif value == [1]:
                # select spin down block
                self.spd = self.spd[:, numofbands:, 0]

        return

    def selectAtoms(self, value=None, fortran=False):
        """
    value is a list with the values of Atoms to select. The optional
    `fortran` argument indicates whether a c-like 0-based indexing
    (`=False`, default) or a fortran-like 1-based (`=True`) is
    provided in `value`.

    Example:
    >>> foo = ProcarParser()
    >>> foo.readFile("PROCAR")
    >>> bar = ProcarSelect(foo)
    >>> bar.selectIspin([...])
    >>> bar.selectAtoms([0,1,2]) #atom0+atom1+atom2

    Note: this method should be called after select.Ispin
    """
        self.log.debug("selectAtoms: ...")

        # taking care about stupid fortran indexing
        if fortran is True:
            value = [x - 1 for x in value]

        # all kpoint, all bands, VALUE atoms, all the rest
        self.log.debug("old shape =" + str(self.spd.shape))

        # testing if the dimensionaluty is rigth:
        dimen = len(self.spd.shape)
        if dimen != 4:
            self.log.error(
                "The array is " + str(dimen) + " dimensional, expecting a"
                " 4 dimensional array."
            )
            self.log.error(
                "You should call selectIspin->selecAtom->selectOrbitals, "
                "in this order."
            )
            raise RuntimeError("Wrong dimensionality of the array")
        self.spd = self.spd[:, :, value]
        self.spd = self.spd.sum(axis=2)
        # self.cspd = self.cspd[:,:,value]
        # self.cspd = self.cspd.sum(axis=2)
        self.log.info("new shape =" + str(self.spd.shape))
        self.log.debug("selectAtoms: ...Done")
        return

    def selectOrbital(self, value):
        """
    value is a list with the values of orbital to select.

    Example:
    >>> foo = ProcarParser()
    >>> foo.readFile("PROCAR")
    >>> bar = ProcarSelect(foo)
    >>> bar.selectIspin([...])
    >>> bar.selectAtoms([...])
    >>> bar.selectOrbital([-1]) #the last (`tot`) field

    to select "p" orbitals just change the argument in the last line
    to [2,3,4] or as needed

    Note: this method should be called after `select.Ispin` and
    `select.Atoms`
    """
        self.log.debug("selectOrbital: ...")
        self.log.debug("Changing the orbital `values` to have a 0-based indexes")
        # Mind: the first orbital field is the atoms number, which is not
        # an orbital, therefore the orbital index is an affective 1-based
        # therefore all `value` indexes += 1 (well, negative values do not
        # change )
        for i in range(len(value)):
            if value[i] >= 0:
                value[i] += 1

        self.log.debug("New values (indexes to select) :" + str(value))

        # all kpoint, all bands, VALUE orbitals, nothing else?
        self.spd = self.spd[:, :, value]
        # self.cspd = self.cspd[:,:,value]
        self.log.debug("old shape =" + str(self.spd.shape))

        # testing if the dimensionaluty is rigth:
        dimen = len(self.spd.shape)
        if dimen != 3:
            self.log.error(
                "The array is " + str(dimen) + " dimensional, expecting a"
                " 3 dimensional array."
            )
            self.log.error(
                "You should call selectIspin->selecAtom->selectOrbitals, "
                "in this order."
            )
            raise RuntimeError("Wrong dimensionality of the array")

        self.spd = self.spd.sum(axis=2)
        # self.cspd=self.cspd.sum(axis=2)
        self.log.info("new shape =" + str(self.spd.shape))
        self.log.debug("selectOrbital: ...Done")
        return



"""
Created on Sunday, May 24th
@author: Logan Lang
"""

from re import findall, search, match, DOTALL, MULTILINE, finditer, compile

from numpy import array, dot, linspace, sum, where, zeros, pi
import logging


class QEParser:
    def __init__(
        self, bandsin="bands.in", kpdosin="kpdos.in", outfile="scf.out", kdirect=True
    ):

        # Qe inputs
        self.bandsin = bandsin
        self.kfin = kpdosin
        self.outfile = outfile

        # This is not used since the extra files are useless for the parser
        # self.file_names = []
        # Not used
        self.kdirect = kdirect
        self.kticks = None
        self.discontinuities = []
        ############################
        self.high_symmetry_points = None
        self.nhigh_sym = None
        self.nspecies = None

        self.composition = {}

        self.kticks = None
        self.knames = None

        self.kpoints = None
        self.kpointsCount = None
        self.bands = None
        self.bandsCount = None
        self.ionsCount = None

        # Used to store atomic states at the top of the kpdos.out file
        self.states = None
        self.test = None
        self.spd = None
        """ for l=1:
              1 pz     (m=0)
              2 px     (real combination of m=+/-1 with cosine)
              3 py     (real combination of m=+/-1 with sine)
            for l=2:
              1 dz2    (m=0)
              2 dzx    (real combination of m=+/-1 with cosine)
              3 dzy    (real combination of m=+/-1 with sine)
              4 dx2-y2 (real combination of m=+/-2 with cosine)
              5 dxy    (real combination of m=+/-2 with sine)"""
        # Oribital order. This is the way it goes into the spd Array. index 0 is resered for totals
        self.orbitals = [
            {"l": 0, "m": 1},
            {"l": 1, "m": 3},
            {"l": 1, "m": 1},
            {"l": 1, "m": 2},
            {"l": 2, "m": 5},
            {"l": 2, "m": 3},
            {"l": 2, "m": 1},
            {"l": 2, "m": 2},
            {"l": 2, "m": 4},
        ]

        # These are not used
        self.orbitalName_short = ["s", "p", "d", "tot"]
        self.orbitalCount = None

        # Opens the needed files
        rf = open(self.bandsin, "r")
        self.bandsIn = rf.read()
        rf.close()

        rf = open(self.kfin, "r")
        self.kpdosIn = rf.read()
        rf.close()

        self.test = None

        # The only method this parser takes. I could make more methods to increase its modularity
        self._readQEin()

        return

    def _readQEin(self):

        ###############################################

        spinCalc = False
        if len(findall("\s*nspin=(.*)", self.bandsIn)) != 0:
            spinCalc = True

        #######################################################################
        # Finding high symmetry points
        #######################################################################

        self.nhigh_sym = int(findall("K_POINTS.*\n([0-9]*)", self.bandsIn)[0])
        raw_khigh_sym = findall(
            "K_POINTS.*\n\s*[0-9]*.*\n" + self.nhigh_sym * "([0-9\s\.-]*).*\n",
            self.bandsIn,
        )[0]
        # self.kpointsCount = 0

        self.kticks = []
        self.high_symmetry_points = zeros(shape=(self.nhigh_sym, 3))
        tick_Count = 1
        for ihs in range(self.nhigh_sym):
            self.high_symmetry_points[ihs, :] = [
                float(x) for x in raw_khigh_sym[ihs].split()[0:3]
            ]
            # kpoint_Count += raw_khigh_sym[ihs].split()[4]
            self.kticks.append(tick_Count - 1)
            tick_Count += int(raw_khigh_sym[ihs].split()[3])

        #######################################################################
        # Finding composition and specie data
        #######################################################################

        self.nspecies = int(findall("ntyp\s*=\s*([0-9]*)", self.bandsIn)[0])
        self.ionsCount = int(findall("nat\s*=\s*([0-9]*)", self.bandsIn)[0])
        raw_species = findall(
            "ATOMIC_SPECIES.*\n" + self.nspecies * "(.*).*\n", self.bandsIn
        )[0]
        species_list = []
        if self.nspecies == 1:
            self.composition[raw_species.split()[0]] = 0
            species_list.append(raw_species.split()[0])
        else:
            for nspec in range(self.nspecies):
                species_list.append(raw_species[nspec].split()[0])
                self.composition[raw_species[nspec].split()[0]] = 0

        raw_ions = findall(
            "ATOMIC_POSITIONS.*\n" + self.ionsCount * "(.*).*\n", self.bandsIn
        )[0]

        if self.ionsCount == 1:
            self.composition[raw_species.split()[0]] = 1
        else:
            for ions in range(self.ionsCount):
                for species in range(len(species_list)):
                    if raw_ions[ions].split()[0] == species_list[species]:
                        self.composition[raw_ions[ions].split()[0]] += 1

        #######################################################################
        # Finding k labels
        #######################################################################

        raw_ticks = findall(
            "K_POINTS.*\n\s*[0-9]*\s*[0-9]*.*\n" + self.nhigh_sym * ".*!(.*)\n",
            self.bandsIn,
        )[0]

        if len(raw_ticks) != self.nhigh_sym:
            self.knames = [str(x) for x in range(self.nhigh_sym)]
        else:
            self.knames = [
                "$%s$" % (x.replace(",", "").replace("vlvp1d", "").replace(" ", ""))
                for x in raw_ticks
            ]

        # note:
        # discontinuities in QE are already specified in the input file.
        # The following is not needed.

        # Fixing discontunuities points if present.
        # Assume that they are separated by a '|'. E.g.- 'G|K'

        # dicontinuity_index = []
        # discontinuity_symbols = [s for s in self.knames if "|" in s]
        # for k in discontinuity_symbols:
        #     dicontinuity_index.append(self.knames.index(k))

        # for i in dicontinuity_index:
        #     self.discontinuities.append(self.kticks[i])

        #######################################################################
        # Reading the kpdos.out for outputfile labels
        #######################################################################
        rf = open(self.kfin.split(".")[0] + ".out", "r")
        kpdosout = rf.read()
        rf.close()

        # The following lines get the number of states in kpdos.out and stores them in a list, in which their information is stored in a dictionary
        raw_natomwfc = findall(
            r"(?<=\(read from pseudopotential files\)\:)[\s\S]*?(?=k)", kpdosout
        )[0]

        natomwfc = len(findall("state[\s#]*(\d*)", raw_natomwfc))

        raw_wfc = findall(
            "\(read from pseudopotential files\)\:.*\n\n\s*" + natomwfc * "(.*)\n",
            kpdosout,
        )[0]

        raw_states = []

        states_list = []
        state_dict = {}

        # Read in raw states
        for state in raw_wfc:
            raw_states.append(
                findall(
                    "state #\s*([0-9]*):\s*atom\s*([0-9]*)\s*\((.*)\s\),\s*wfc\s*([0-9]*)\s\(l=\s*([0-9])\s*m=\s*([0-9])\)",
                    state,
                )[0]
            )
        # self.test = raw_states
        for state in raw_states:
            state_dict = {}
            state_dict = {
                "state_num": int(state[0]),
                "species_num": int(state[1]),
                "specie": state[2],
                "atm_wfc": int(state[3]),
                "l": int(state[4]),
                "m": int(state[5]),
            }
            states_list.append(state_dict)

        self.states = states_list

        """
        This section was used to parse those extra files we discussed. It is not needed, I kept it just incase
        output_prefix = findall("filpdos\s*=\s*'(.*)'", self.kpdosIn)[0]
         kpdos_files=[]
        atmNum = list(range(1, self.ionsCount+1))
        wfc_file_label = []
        atm_file_label = []
        states_dict ={}
        #Find unique raw file labels
        for state1 in raw_states:
            wfc_file_label.append((state1[3],state1[4])) if (state1[3],state1[4]) not in wfc_file_label else None
            atm_file_label.append((state1[1],state1[2])) if (state1[1],state1[2]) not in atm_file_label else None
        #Combine raw file labels inside a tuple
        for atm in atm_file_label:
            for wfc in wfc_file_label:
                file = (int(atm[0]),atm[1],int(wfc[0]),int(wfc[1]))
                kpdos_files.append(file)
                file = None
        #Convert raw file labels to final file string
        for file in kpdos_files:
            if(file[3] == 0 ):
                self.file_names.append(output_prefix + ".pdos_atm#" + str(file[0]) + "(" + file[1] + ")_wfc#" + str(file[2]) +  "(s)")
            elif (file[3] == 1 ):
                self.file_names.append(output_prefix + ".pdos_atm#" + str(file[0]) + "(" + file[1] + ")_wfc#" + str(file[2]) +  "(p)")
            elif (file[3] == 2 ):
                self.file_names.append(output_prefix + ".pdos_atm#" + str(file[0]) + "(" + file[1] + ")_wfc#" + str(file[2]) +  "(d)")
            elif (file[3] == 3 ):
                self.file_names.append(output_prefix + ".pdos_atm#" + str(file[0]) + "(" + file[1] + ")_wfc#" + str(file[2]) +  "(f)")"""

        #######################################################################
        # Kpoints and kpoints index
        #######################################################################

        # Finds the k points and stores them in an array
        raw_kpoints = []
        if spinCalc == True:

            self.kpointsCount = int(
                len(findall("k\s*=\s*(.*)\s*(.*)\s*(.*)", kpdosout)) / 2
            )
            for k in range(len(findall("k\s*=\s*(.*)\s*(.*)\s*(.*)", kpdosout))):
                if k < self.kpointsCount:
                    raw_kpoints.append(
                        findall("k\s*=\s*(.*)\s*(.*)\s*(.*)", kpdosout)[k][0]
                    )
            totK = len(findall("k\s*=\s*(.*)\s*(.*)\s*(.*)", kpdosout))
        else:
            for k in range(len(findall("k\s*=\s*(.*)\s*(.*)\s*(.*)", kpdosout))):
                raw_kpoints.append(
                    findall("k\s*=\s*(.*)\s*(.*)\s*(.*)", kpdosout)[k][0]
                )
            self.kpointsCount = len(raw_kpoints)

        self.kpoints = zeros(shape=(self.kpointsCount, 3))
        for ik in range(len(raw_kpoints)):
            for coord in range(3):
                self.kpoints[ik][coord] = raw_kpoints[ik].split()[coord]

        # If kdirect=False, then the kgrid will be in cartesian coordinates.
        # Requires the reciprocal lattice vectors to be parsed from the output.
        if not self.kdirect:
            self.kpoints = dot(self.kpoints, self.reclat)

        #######################################################################
        # Finds the band count and makes a band array
        #######################################################################

        band_info = findall(r"====\se\(\s*(\d+)\)\s=\s*([-.\d]+)", kpdosout)
        self.test = band_info
        if spinCalc == True:
            self.bandsCount = int(len(band_info) / totK)
            self.bands = zeros(shape=(self.kpointsCount, self.bandsCount, 2))
            ik = 0
            for band in range(len(band_info)):
                if ik < self.kpointsCount:
                    self.bands[ik, int(band_info[band][0]) - 1, 0] = float(
                        band_info[band][1]
                    )
                else:
                    self.bands[
                        ik - self.kpointsCount, int(band_info[band][0]) - 1, 1
                    ] = float(band_info[band][1])
                if int(band_info[band][0]) == self.bandsCount:
                    ik += 1
        else:
            self.bandsCount = int(len(band_info) / self.kpointsCount)
            self.bands = zeros(shape=(self.kpointsCount, self.bandsCount))
            ik = 0
            for band in range(len(band_info)):
                self.bands[ik, int(band_info[band][0]) - 1] = float(band_info[band][1])
                if int(band_info[band][0]) == self.bandsCount:
                    ik += 1

        #######################################################################
        # Filling the spd array
        #######################################################################
        if spinCalc == True:
            spinCount = 2
        else:
            spinCount = 1
        k_info = None
        self.orbitalCount = len(self.orbitals)
        self.spd = zeros(
            shape=(
                self.kpointsCount,
                self.bandsCount,
                spinCount,
                self.ionsCount + 1,
                len(self.orbitals) + 2,
            )
        )

        # Finds the kpoints

        k_string = findall(r"\s?k =[\s-]{3}?.+", kpdosout)

        """First loop goes through the kpoints and matches band information that follows the specific k point. The if else statment is used to catch unique cases, namely the end of the kpoints
           Also sometimes there is a duplicate final k point so findall is used twice to deal with this case
           Second loop goes through band information of a kpoint.The if else statments here again catches unique cases. This is when it hits the last the band or when it is the last band and the last kpoint
           The other if else statement here catches the cases when there are no contributions at all and would return nothing
           Third loop parses the projections of a band
           The fourth loop then goes through the known possible states. The if statment ensures the there is a projection and matchs a projection to a known state.
           The fifth loop then goes throguh the known possible orbitals. The if statment then matches the projection with a specific orbital.
           Finally the projection is put into the spd array
        """

        for kp in range(len(k_string)):
            if kp == len(k_string) - 1:
                expression = "(?<=" + k_string[kp] + ")[\s\S]*?(?=Lowdin Charges:)"
                expression_final_k = "(?<=" + k_string[kp - 1] + ")[\s\S]*?(?=$)"
                k_info1 = findall(expression_final_k, kpdosout)[0]
                k_info = findall(expression, k_info1)[0]
            else:
                expression = (
                    "(?<=" + k_string[kp] + ")[\s\S]*?(?=" + k_string[kp + 1] + ")"
                )
                # print(expression)
                k_info = findall(expression, kpdosout)[0]
                # print(k_info)
                # print(k_info)
                # print(self.bandsCount)
            for iband in range(self.bandsCount):
                if iband == self.bandsCount - 1 and kp == len(k_string) - 1:
                    # print("hi")
                    expression2 = "==== e\(\s*" + str(iband + 1) + "\)[\s\S]*?(?=$)"
                    # print(k_info)
                    iband_proj = findall(expression2, k_info)[0]
                    # print(iband_proj)
                elif iband == self.bandsCount - 1:
                    expression2 = "==== e\(\s*" + str(iband + 1) + "\).*\n([\s\S]*)"
                    iband_proj = findall(expression2, k_info)[0]

                else:
                    # print("hi")
                    expression2 = (
                        "==== e\(\s*"
                        + str(iband + 1)
                        + "\)[\s\S]*?(?===== e\(\s*"
                        + str(iband + 2)
                        + "\))"
                    )
                    iband_proj = findall(expression2, k_info)[0]
                    # print(iband_proj)
                expression3 = "(?<=psi = )[\s\S]*?(?=\s*\|psi\|\^2)"
                wfc = None

                if len(findall(expression3, iband_proj)) == 0:
                    pass
                else:
                    wfc = findall(expression3, iband_proj)[0]
                    # print(wfc)
                    state_proj = wfc.split("+")

                    for proj in state_proj:
                        for known_states in self.states:
                            if (
                                len(findall("#\s*([0-9]*)", proj)) != 0
                                and int(findall("#\s*([0-9]*)", proj)[0])
                                == known_states["state_num"]
                            ):
                                for iorbitals in range(len(self.orbitals)):
                                    if (
                                        known_states["l"]
                                        == self.orbitals[iorbitals]["l"]
                                        and known_states["m"]
                                        == self.orbitals[iorbitals]["m"]
                                    ):
                                        if spinCalc == True:
                                            if kp < self.kpointsCount:
                                                # print(known_states["species_num"])
                                                self.spd[
                                                    kp,
                                                    iband,
                                                    0,
                                                    known_states["species_num"] - 1,
                                                    iorbitals + 1,
                                                ] += float(proj.split("*")[0])
                                                self.spd[
                                                    kp,
                                                    iband,
                                                    0,
                                                    known_states["species_num"] - 1,
                                                    0,
                                                ] = known_states["species_num"]
                                            else:
                                                self.spd[
                                                    kp - self.kpointsCount,
                                                    iband,
                                                    1,
                                                    known_states["species_num"] - 1,
                                                    iorbitals + 1,
                                                ] += float(proj.split("*")[0])
                                                self.spd[
                                                    kp - self.kpointsCount,
                                                    iband,
                                                    1,
                                                    known_states["species_num"] - 1,
                                                    0,
                                                ] = known_states["species_num"]

                                        else:
                                            self.spd[
                                                kp,
                                                iband,
                                                0,
                                                known_states["species_num"] - 1,
                                                iorbitals + 1,
                                            ] += float(proj.split("*")[0])
                                            self.spd[
                                                kp,
                                                iband,
                                                0,
                                                known_states["species_num"] - 1,
                                                0,
                                            ] = known_states["species_num"]

        for ions in range(self.ionsCount):
            self.spd[:, :, :, ions, 0] = ions + 1

        # The following fills the totals for the spd array

        self.spd[:, :, :, :, -1] = sum(self.spd[:, :, :, :, 1:-1], axis=4)
        self.spd[:, :, :, -1, :] = self.spd.sum(axis=3)
        self.spd[:, :, :, -1, 0] = 0

        # colinear spin polarized case

        # manipulating spd array for spin polarized calculations.
        # The shape is (nkpoints,2*nbands,2,natoms,norbitals)
        # The third dimension is for spin.
        # When this is zero, the bands*2 (all spin up and down bands) have positive projections.
        # When this is one, the the first half of bands (spin up) will have positive projections
        # and the second half (spin down) will have negative projections. This is to adhere to
        # the convention used in PyProcar to obtain spin density and spin magnetization.

        if spinCalc:
            print("\nQuantum Espresso colinear spin calculation detected.\n")
            self.spd2 = zeros(
                shape=(
                    self.kpointsCount,
                    self.bandsCount * 2,
                    spinCount,
                    self.ionsCount + 1,
                    len(self.orbitals) + 2,
                )
            )

            # spin up block for spin=0
            self.spd2[:, : self.bandsCount, 0, :, :] = self.spd[:, :, 0, :, :]

            # spin down block for spin=0
            self.spd2[:, self.bandsCount :, 0, :, :] = self.spd[:, :, 1, :, :]

            # spin up block for spin=1
            self.spd2[:, : self.bandsCount, 1, :, :] = self.spd[:, :, 0, :, :]

            # spin down block for spin=1
            self.spd2[:, self.bandsCount :, 1, :, :] = -1 * self.spd[:, :, 1, :, :]

            self.spd = self.spd2

    @property
    def fermi(self):
        """
        Returns the fermi energy read from .out
        """

        fi = open(self.outfile, "r")
        data = fi.read()
        fi.close()
        fermi = float(findall(r"the\s*Fermi\s*energy\s*is\s*([\s\d.]*)ev", data)[0])
        return fermi

    @property
    def reclat(self):
        """
        Returns the reciprocal lattice read from .out
        """
        rf = open(self.outfile, "r")
        data = rf.read()
        rf.close()

        alat = float(findall(r"alat\)\s*=\s*([\d.e+-]*)", data)[0])

        b1 = array(
            findall(r"b\(1\)\s*=\s*\(([\d.\s+-e]*)", data)[0].split(), dtype="float64"
        )
        b2 = array(
            findall(r"b\(2\)\s*=\s*\(([\d.\s+-e]*)", data)[0].split(), dtype="float64"
        )
        b3 = array(
            findall(r"b\(3\)\s*=\s*\(([\d.\s+-e]*)", data)[0].split(), dtype="float64"
        )

        reclat = (2 * pi / alat) * (array((b1, b2, b3)))

        # Transposing to get the correct format
        reclat = reclat.T

        return reclat
